#
# CellML 1.0/1.1 API
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import collections
import re
import warnings

import myokit


# Identifier validation
_cellml_identifier = re.compile('^([_][0-9_]*)?[a-zA-Z][a-zA-Z0-9_]*$')


def is_valid_identifier(name):
    """
    Tests if the given ``name`` is a valid CellML 1.1 identifier.

    This method returns True if (and only if) the identifier

    1. contains only alphanumeric characters (from the basic latin set) or
       underscores
    2. contains at least one letter
    3. does not begin with a numerical character.

    The rules for 1.0 are slightly more lenient, and allow silly things like
    ``1e2``, ``123`` or ``_123`` as identifiers. Because this creates issues
    distinguishing identifiers from numbers, Myokit always follows the 1.1
    rule, even for 1.0 documents.
    """
    return _cellml_identifier.match(name) is not None


def clean_identifier(name):
    """
    Checks if ``name`` is a valid CellML identifier and if not attempts to make
    it into one.

    Raises a ``ValueError`` if it can't create a valid identifier.
    """
    if is_valid_identifier(name):
        return name

    # Replace spaces and hyphens with underscores
    clean = re.sub(r'[\s-]', '_', name)

    # Check if valid and return
    if is_valid_identifier(clean):
        return clean
    raise ValueError(
        'Unable to create a valid CellML identifier from "' + str(name) + '".')


def create_unit_name(unit):
    """
    Creates an almost readable name for a Myokit ``unit``.
    """
    # Get preferred name from Myokit's representation (e.g. [kg]) but trim off
    # the brackets
    name = str(unit)[1:-1]

    # If this is a valid name, return
    if is_valid_identifier(name):
        return name

    # Not allowed: could be because of a multiplier, e.g. [m (0.0254)]
    if ' ' in name:
        name, multiplier = name.split(' ')
    else:
        name, multiplier = name, ''

    # Could also be because it's g*m/mol^2
    name = name.replace('^', '')
    name = name.replace('/', '_per_')
    name = name.replace('*', '_')
    name = name.replace('.', '_dot_')

    # Remove "1_" from "1_per_mV"
    if name[:2] == '1_':
        name = name[2:]

    # Turn "1 (123)" into "dimensionless (123)"
    elif name == '1':
        # E.g. [1 (1000)]
        name = 'dimensionless'

    # Add multiplier (can be float)
    if multiplier:
        multiplier = unit.multiplier()

        # Use e-notation if multiple of 10
        multiplier10 = unit.multiplier_log_10()
        if myokit.float.eq(multiplier10, int(multiplier10)):
            multiplier = '1e' + str(int(multiplier10))

        # Format as integer
        elif myokit.float.eq(multiplier, int(multiplier)):
            multiplier = str(int(multiplier))

        # Format as float
        else:
            multiplier = str(multiplier)

        # Remove characters not allowed in CellML identifiers
        multiplier = multiplier.replace('+', '')
        multiplier = multiplier.replace('-', '_minus_')
        multiplier = multiplier.replace('.', '_dot_')
        name += '_times_' + multiplier

    return name


class AnnotatableElement:
    """
    Represents a CellML 1.0 or 1.1 element that can have a cmeta:id.

    Each ``AnnotatableElement`` also has a public dict ``meta`` that can be
    used to story key:value (string:string) pairs, for storing meta data.
    """

    def __init__(self, model):

        # The model this element is in (or is)
        self._model = model

        # A unique identifier (or None)
        self._cmeta_id = None

        # Public meta data
        self.meta = {}

    def cmeta_id(self):
        """
        Returns this element's cmeta:id if set, or ``None`` otherwise.
        """
        return self._cmeta_id

    def set_cmeta_id(self, cmeta_id):
        """
        Sets this element's cmeta id (must be a non-empty string or ``None``).
        """
        # Check if already set, ignore setting twice
        if self._cmeta_id is not None:
            # Ignore setting the same id again
            if cmeta_id == self._cmeta_id:
                return

        # Set new cmeta id, unregister old one
        if cmeta_id is None:
            if self._cmeta_id is not None:
                self._model._unregister_cmeta_id(self._cmeta_id)
            self._cmeta_id = None
        else:
            # Register and check uniqueness etc.
            self._model._register_cmeta_id(cmeta_id, self)

            # Unregister old (do _after_ register has checked new cmeta id)
            if self._cmeta_id is not None:
                self._model._unregister_cmeta_id(self._cmeta_id)

            # Store new
            self._cmeta_id = cmeta_id


class CellMLError(myokit.MyokitError):
    """
    Raised when an invalid CellML model is created or detected, or when a model
    uses CellML features that Myokit does not support.
    """


class UnitsError(CellMLError):
    """
    Raised when unsupported unit features are used.
    """


class UnsupportedBaseUnitsError(UnitsError):
    """
    Raised when unsupported base units are used.
    """
    def __init__(self, units):
        self.units = units
        super().__init__(
            'Unsupported base units "' + units + '".')


class UnsupportedUnitOffsetError(UnitsError):
    """
    Raised when units with non-zero offsets are used.
    """
    def __init__(self):
        super().__init__(
            'Units with non-zero offsets are not supported.')


class Component(AnnotatableElement):
    """
    Represents a model component, should not be created directly but only via
    :meth:`Model.add_component()`.
    """
    def __init__(self, model, name):
        super().__init__(model)

        # Store model
        self._model = model

        # Check and store name
        if not is_valid_identifier(name):
            raise CellMLError(
                'Component name must be a valid CellML identifier (3.4.2.2).')
        self._name = name

        # This component's encapsulation relationships
        self._parent = None
        self._children = set()

        # This component's variables
        self._variables = collections.OrderedDict()

        # This component's component-level units
        self._units = {}

        # Lookup myokit unit to name (note this lookup may not be unique)
        self._myokit_unit_to_name = {}

    def add_units(self, name, myokit_unit):
        """
        Add a component-level units definition to this component.

        Arguments:

        ``name``
            A valid CellML identifier to use as the name
        ``myokit_unit``
            A :class:`myokit.Unit` object.
        """
        # Check uniqueness
        if name in self._units:
            raise CellMLError(
                'Units defined in the same component cannot have the same'
                ' name (5.4.1.2)')

        # Create, store, and return
        # Note: there can be units with different names but the same myokit
        # unit, in this case we simply overwrite whatever name was already
        # there. This shouldn't cause any issues because the name still refers
        # to a valid unit.
        units = Units(name, myokit_unit)
        self._units[name] = units
        self._myokit_unit_to_name[myokit_unit] = name
        return units

    def add_variable(self, name, units, public_interface='none',
                     private_interface='none'):
        """
        Adds a variable with the given ``name`` and ``units``.

        Arguments

        ``name``
            A valid CellML identifier (string).
        ``units``
            The name of a units definition known to this component or its
            parent model.
        ``public_interface``
            The variable's public interface.
        ``private_interface``
            The variable's private interface.

        """
        # Check name uniqueness
        if name in self._variables:
            raise CellMLError(
                'Variable name must be unique within Component (3.4.3.2).')

        # Create, store, and return
        self._variables[name] = v = Variable(
            self, name, units, public_interface, private_interface)
        return v

    def children(self):
        """
        Returns an iterator over any encapsulated child components.
        """
        return iter(self._children)

    def __contains__(self, key):
        return key in self._variables

    def find_units(self, name):
        """
        Looks up and returns a :class:`Units` object with the given ``name``.

        Searches first in this component, then in its model, then in the list
        of predefined units.

        Raises a :class:`CellMLError` is no unit is found.
        """
        try:
            return self._units[name]
        except KeyError:
            return self._model.find_units(name)

    def find_units_name(self, myokit_unit):
        """
        Attempts to find a string name for the given :class:`myokit.Unit`.

        Searches first in this component, then in its model, then in the list
        of predefined units. If multiple units definitions have the same
        :class:`myokit.Unit`, the last added name is returned.

        Raises a :class:`CellMLError` is no appropriate unit is found.
        """
        try:
            return self._myokit_unit_to_name[myokit_unit]
        except KeyError:
            return self._model.find_units_name(myokit_unit)

    def __getitem__(self, key):
        return self._variables[key]

    def has_children(self):
        """
        Checks if this component has any encapsulated child components.
        """
        return len(self._children) > 0

    def __iter__(self):
        return iter(self._variables.values())

    def __len__(self):
        return len(self._variables)

    def model(self):
        """
        Return this component's model.
        """
        return self._model

    def name(self):
        """
        Returns this component's name.
        """
        return self._name

    def parent(self):
        """
        Returns this component's parent component (or None).
        """
        return self._parent

    def set_parent(self, parent):
        """
        Sets this component as the encapsulated child of the component
        ``parent``.
        """
        # Check parent is a component from the same model (or None)
        if parent is not None:
            if not isinstance(parent, Component):
                raise ValueError('Parent must be a cellml.v1.Component.')
            if self._model != parent._model:
                raise ValueError('Parent must be from the same model.')

        # Store relationship
        if self._parent is not None:
            self._parent._children.remove(self)
        self._parent = parent
        if parent is not None:
            parent._children.add(self)

    def __str__(self):
        return 'Component[@name="' + self._name + '"]'

    def units(self):
        """
        Returns an iterator over the :class:`Units` objects in this component.
        """
        return iter(self._units.values())

    def _validate(self):
        """
        Validates this component, raising a :class:`CellMLError` if any errors
        are found.
        """
        # Check for cycles in the encapsulation hierarchy
        parent = self._parent
        while parent is not None:
            if parent is self:
                raise CellMLError(
                    'Encapsulation hierarchy cannot be circular (6.4.3.2).')
            parent = parent._parent

        # Validate variables
        for v in self._variables.values():
            v._validate()

        # If this component has states, check that it also has a free variable
        # in the local component.
        has_states = False
        has_free = False
        for v in self._variables.values():
            if v.is_state():
                has_states = True
            elif v.value_source().is_free():
                has_free = True
            if has_states and has_free:
                break

        # If derivatives are used, check that a time variable is defined and
        # is available in this component.
        if has_states:
            if self._model.free_variable() is None:
                raise CellMLError(
                    'If state variables are used, a free variable must be set'
                    ' with Model.set_free_variable().')
            if not has_free:
                raise CellMLError(
                    str(self) + ' has state variables, but no local variable'
                    ' connected to the free variable.')

    def variable(self, name):
        """
        Returns the variable with the given ``name``.
        """
        return self._variables[name]

    def variables(self):
        """
        Returns an iterator over this component's variables.
        """
        return iter(self._variables.values())


class Model(AnnotatableElement):
    """
    Represents a CellML 1.0 or 1.1 model.

    Support notes for 1.0 and 1.1:

    - Units with offsets are not supported, including the base units "celsius".
    - Defining new base units is not supported.
    - Reactions are not supported.
    - Models that take derivatives with respect to more than one variable are
      not supported.
    - Models written as a DAE (e.g. ``1 + x = 2 + y``) are not supported.
    - cmeta:id support is limited to models, components, and variables.

    Support notes for 1.1:

    - The new feature of using variables in ``initial_value`` attributes is not
      supported.
    - Imports (CellML 1.1) are not supported.

    Support notes for 1.0:

    - The stricter 1.1 rule for identifiers is used for both CellML 1.0 and
      1.1: a valid identifier not start with a number, and must contain at
      least one letter.

    Arguments:

    ``name``
        A valid CellML identifier string.
    ``version``
        A string representing the CellML version this model is in (must be
        '1.0' or '1.1').

    """
    def __init__(self, name, version='1.0'):
        super().__init__(self)

        # Check and store name
        if not is_valid_identifier(name):
            raise CellMLError(
                'Model name must be a valid CellML identifier (3.4.1.2).')
        self._name = name

        # Check and store version
        if version not in ('1.0', '1.1'):
            raise ValueError(
                'Only 1.0 and 1.1 models are supported by this API.')
        self._version = version

        # This model's components
        self._components = collections.OrderedDict()

        # This model's model-level units
        self._units = {}

        # Lookup myokit unit to name (note this lookup may not be unique)
        self._myokit_unit_to_name = {}

        # A mapping from cmeta:ids to CellML element objects.
        self._cmeta_ids = {}

        # This model's free variable
        self._free_variable = None

        # Note: encapsulation relationships are stored inside the components,
        # other relationship types are not stored at all

        # Note: connections are stored in the variable with the "in" interface.

    def add_component(self, name):
        """
        Adds an empty component with the given ``name``.
        """
        # Check uniqueness
        if name in self._components:
            raise CellMLError(
                'Component name must be unique within model (3.4.2.2).')

        # Create, store, and return
        self._components[name] = c = Component(self, name)
        return c

    def add_connection(self, variable_1, variable_2):
        """
        Adds a connection between ``variable_1`` and ``variable_2``.
        """
        # Check both are variables, and from this model
        if not isinstance(variable_1, Variable):
            raise ValueError('Argument variable_1 must be a'
                             ' cellml.v1.Variable.')
        if not isinstance(variable_2, Variable):
            raise ValueError('Argument variable_2 must be a'
                             ' cellml.v1.Variable.')
        if variable_1._model is not self:
            raise ValueError('Argument variable_1 must be a variable from this'
                             ' model.')
        if variable_2._model is not self:
            raise ValueError('Argument variable_2 must be a variable from this'
                             ' model.')

        # Check variables are distinct
        if variable_1 is variable_2:
            raise CellMLError(
                'Variables cannot be connected to themselves.')

        # Check components are distinct
        component_1 = variable_1.component()
        component_2 = variable_2.component()
        if component_1 is component_2:
            raise CellMLError('Variables cannot be connected to variables in'
                              ' the same component (3.4.5.4)')

        # Determine which interfaces connect these variables
        interface_1 = variable_1.public_interface()
        interface_2 = variable_2.public_interface()
        string_1 = string_2 = 'public'
        if component_1 == component_2.parent():
            # Component 1 is the parent and uses the private interface
            interface_1 = variable_1.private_interface()
            string_1 = 'private'
        elif component_2 == component_1.parent():
            # Component 2 is the parent and uses the private interface
            interface_2 = variable_2.private_interface()
            string_2 = 'private'
        elif component_1.parent() != component_2.parent():
            raise CellMLError(
                'Connections can only be made between components that are'
                ' siblings or have a parent-child (encapsulation)'
                ' relationship (3.4.6.4).')

        # Check interfaces and connect variables
        # Note: We're allowing the same connection to be specified twice here
        if interface_1 == 'out' and interface_2 == 'in':
            if variable_2._source is None:
                variable_2._source = variable_1
            elif variable_2._source is variable_1:
                raise CellMLError(
                    'Invalid connection: ' + str(variable_2) + ' is already'
                    ' connected to ' + str(variable_1) + '.')
            else:
                raise CellMLError(
                    'Invalid connection: ' + str(variable_2) + ' has a '
                    + string_1 + '_interface of "in" and is already connected'
                    ' to a variable with an interface of "out".')
        elif interface_1 == 'in' and interface_2 == 'out':
            if variable_1._source is None:
                variable_1._source = variable_2
            elif variable_1._source is variable_2:
                raise CellMLError(
                    'Invalid connection: ' + str(variable_1) + ' is already'
                    ' connected to ' + str(variable_2) + '.')
            else:
                raise CellMLError(
                    'Invalid connection: ' + str(variable_1) + ' has a '
                    + string_2 + '_interface of "in" and is already connected'
                    ' to a variable with an interface of "out".')
        else:
            raise CellMLError(
                'Invalid connection: ' + str(variable_1) + ' has a ' + string_1
                + '_interface of "' + interface_1 + '", while '
                + str(variable_2) + ' has a ' + string_2 + '_interface of "'
                + interface_2 + '" (3.4.6.4).')

    def add_units(self, name, myokit_unit):
        """
        Add a model-level units definition to this model.

        Arguments:

        ``name``
            A valid CellML identifier to use as the name
        ``myokit_unit``
            A :class:`myokit.Unit` object.
        """
        # Check uniqueness
        if name in self._units:
            raise CellMLError(
                'Units defined in the same model cannot have the same name'
                ' (5.4.1.2)')

        # Create, store, and return
        units = Units(name, myokit_unit)
        self._units[name] = units
        self._myokit_unit_to_name[myokit_unit] = name
        return units

    def component(self, name):
        """
        Returns the :class:`Component` with the given ``name``.
        """
        return self._components[name]

    def components(self):
        """
        Returns an iterator over this model's components.
        """
        return iter(self._components.values())

    def __contains__(self, key):
        return key in self._components

    def element_with_cmeta_id(self, cmeta_id):
        """
        Returns the model, component, or variable with the given ``cmeta_id``
        or raises a KeyError if no such element is found.
        """
        return self._cmeta_ids[cmeta_id]

    def find_units(self, name):
        """
        Looks up and returns a :class:`Units` object with the given ``name``.

        Searches first in this model, then in the list of predefined units.

        Raises a :class:`CellMLError` is no unit is found.
        """
        try:
            return self._units[name]
        except KeyError:
            return Units.find_units(name)

    def find_units_name(self, myokit_unit):
        """
        Attempts to find a string name for the given :class:`myokit.Unit`.

        Searches first in this model, then in the list of predefined units. If
        multiple units definitions have the same :class:`myokit.Unit`, the last
        added name is returned.

        Raises a :class:`CellMLError` is no appropriate unit is found.
        """
        try:
            return self._myokit_unit_to_name[myokit_unit]
        except KeyError:
            return Units.find_units_name(myokit_unit)

    def free_variable(self):
        """
        Returns the free variable set with :meth:`set_free_variable`.
        """
        return self._free_variable

    @staticmethod
    def from_myokit_model(model, version='1.0'):
        """
        Creates a CellML :class:`Model` from a :class:`myokit.Model`.

        The CellML version to use can be set with ``version``, which must be
        either "1.0" or "1.1".
        """
        # Model must be valid
        # Otherwise could have cycles, invalid references, etc.
        model.validate()

        # Get name for CellML model
        name = model.name()
        if name is None:
            name = 'unnamed_myokit_model'
        else:
            try:
                name = clean_identifier(name)
            except ValueError:
                name = 'unnamed_myokit_model'

        # Create CellML model
        m = Model(name, version)

        # Valid model always has a time variable
        time = model.time()

        # Method to obtain or infer variable unit
        def variable_unit(variable, time_unit):
            """Returns variable.unit(), or attempts to infer it if not set."""
            unit = variable.unit()
            if unit is not None:
                return unit

            rhs = variable.rhs()
            if rhs is not None:
                try:
                    # Tolerant evaluation, see above. Result may be None.
                    unit = rhs.eval_unit(myokit.UNIT_TOLERANT)
                except myokit.IncompatibleUnitError:
                    return None
                if variable.is_state():
                    if unit is not None and time_unit is not None:
                        # RHS is divided by time unit, so multiply
                        unit *= time_unit
            return unit

        # Time unit, used to infer units of state variables
        # (May itself be inferred)
        time_unit = variable_unit(time, None)

        # Gather unit objects used in Myokit model, and gather numbers without
        # units that will need to be replaced.
        used = set()
        numbers_without_units = {}
        for variable in model.variables(deep=True):

            # Add variable unit, or unit inferred from variable's RHS
            used.add(variable_unit(variable, time_unit))

            # Check number units
            rhs = variable.rhs()
            if rhs is not None:
                for e in rhs.walk(myokit.Number):
                    u = e.unit()
                    if u is None:
                        numbers_without_units[e] = myokit.Number(
                            e.eval(), myokit.units.dimensionless)
                    used.add(u)

        # Remove 'None' from the set of used units
        if None in used:
            used.remove(None)

        # Add units to models, and store mapping from objects to names
        unit_map = dict(Units._si_units_r)
        for unit in used:
            if unit not in unit_map:
                name = create_unit_name(unit)
                unit_map[unit] = name
                m.add_units(name, unit)

        # Map None to dimensionless
        unit_map[None] = 'dimensionless'

        # Variable naming strategy:
        # 1. Component level variables always use their unqualified names
        # 2. Nested variables use their unames
        # 3. Variable-in variables use their unames

        # Create unames in model
        model.create_unique_names()

        # Dict of interface-in variables that need to be added to each
        # component
        in_variables = {component: set() for component in model}

        # Dict mapping Myokit variables to CellML variables, per component
        var_map = {component: dict() for component in model}

        # Add components
        for component in model:
            c = m.add_component(component.name())

            # Local var_map entry
            local_var_map = var_map[component]

            # Add variables
            for variable in component:

                # Check if this variable is needed in other components
                interface = 'none'

                # Get all refs to variable's LHS
                refs = set(variable.refs_by())
                if variable.is_state():
                    # If it's a state, refs to dot(x) also require the time
                    # variable
                    if refs:
                        refs.add(time)

                    # And refs should include references to the state itself
                    refs = refs.union(variable.refs_by(True))

                elif variable is time:
                    # If this is the time variable, then ensure that all states
                    # have a local reference to it.
                    for ref in model.states():
                        refs.add(ref)

                # Update lists of required interface-in variables, and set
                # interface to 'out' if needed
                for user in refs:
                    parent = user.parent(myokit.Component)
                    if parent != component:
                        in_variables[parent].add(variable)
                        interface = 'out'

                # Get declared or inferred variable unit
                unit = variable_unit(variable, time_unit)

                # Add variable
                local_var_map[variable] = v = c.add_variable(
                    variable.name(),
                    unit_map[unit],
                    public_interface=interface
                )

                # Copy meta data
                for key, value in variable.meta.items():
                    v.meta[key] = value

                # Create cmeta id for variables with an oxmeta annotation
                if 'oxmeta' in variable.meta:
                    v.set_cmeta_id(variable.uname())

                # Add nested variables
                for nested in variable.variables(deep=True):
                    local_var_map[nested] = v = c.add_variable(
                        nested.uname(),
                        unit_map[variable_unit(nested, time_unit)])

                    # Copy meta-data
                    for key, value in nested.meta.items():
                        v.meta[key] = value

                    # Create cmeta id for variables with an oxmeta annotation
                    if 'oxmeta' in nested.meta:
                        v.set_cmeta_id(nested.uname())

        # Add interface-in variables
        for component, variables in in_variables.items():
            c = m[component.name()]

            # Local var_map entry
            local_var_map = var_map[component]

            # Add variables
            for variable in variables:
                local_var_map[variable] = v = c.add_variable(
                    variable.uname(),
                    unit_map[variable_unit(variable, time_unit)],
                    public_interface='in',
                )

                # Add connection
                source = m[variable.parent().name()][variable.name()]
                m.add_connection(v, source)

        # Set RHS equations
        for component in model:

            # Create dict of Name substitutions
            local_var_map = var_map[component]
            subst = {
                myokit.Name(x): myokit.Name(y)
                for x, y in local_var_map.items()}

            # Add number substitutions
            for x, y in numbers_without_units.items():
                subst[x] = y

            # Set RHS equations
            for variable in component.variables(deep=True):
                v = local_var_map[variable]

                # Create RHS with updated numbers and references
                rhs = variable.rhs().clone(subst=subst)

                # Free variable shouldn't have a value
                if variable is time:
                    m.set_free_variable(v)

                # Promote states and set rhs and initial value
                elif variable.is_state():
                    v.set_is_state(True)
                    v.set_initial_value(variable.initial_value(True))
                    v.set_rhs(rhs)

                # Store literals (single number) in initial value
                elif isinstance(rhs, myokit.Number):
                    v.set_initial_value(rhs.eval())

                # For all other use rhs
                else:
                    v.set_rhs(rhs)

        # Return model
        return m

    def __getitem__(self, key):
        return self._components[key]

    def __iter__(self):
        return iter(self._components.values())

    def __len__(self):
        return len(self._components)

    def myokit_model(self):
        """
        Returns this model's :class:`myokit.Model` equivalent.
        """
        # Make sure model is valid before starting
        self.validate()

        # Create model
        m = myokit.Model(self.name())
        m.meta['author'] = 'Myokit CellML 1 API'

        # Copy meta data
        for key, value in self.meta.items():
            m.meta[key] = value

        # Gather set of variables that are used in (local) equations
        used = set()
        for component in self:
            for variable in component:
                rhs = variable.rhs()
                if rhs is not None:
                    for ref in rhs.references():
                        used.add(ref.var())

        # Gather dict of variables that need unit conversion, mapping their
        # CellML variables to a tuple of units (from, to).
        # This is the set of variables that are non-local, have a different
        # unit than their source, and are referenced by other variables.
        needs_conversion = set()
        for component in self:
            for variable in component:
                if variable.is_local() or variable not in used:
                    continue

                # Compare units
                ufrom = variable.value_source().units().myokit_unit()
                uto = variable.units().myokit_unit()
                if ufrom != uto:
                    needs_conversion.add(variable)

        # Add components
        for component in self:
            c = m.add_component(component.name())

            # Copy meta data
            for key, value in component.meta.items():
                c.meta[key] = value

            # Create local variables or variables that need unit conversion
            for variable in component:
                if variable.is_local() or variable in needs_conversion:
                    v = c.add_variable(variable.name())
                    v.set_unit(variable.units().myokit_unit())

                    # Copy meta data
                    for key, value in variable.meta.items():
                        v.meta[key] = value

        # Add equations
        undefined_variables = []
        for component in self:
            c = m[component.name()]

            # Create dict of variable reference substitutions
            var_map = {}
            for variable in component:
                cname = myokit.Name(variable)
                if variable.is_local() or variable in needs_conversion:
                    mname = myokit.Name(c[variable.name()])
                else:
                    source = variable.value_source()
                    source = m[source.component().name()][source.name()]
                    mname = myokit.Name(source)
                var_map[cname] = mname

            # Add equations
            for variable in component:

                # Add local variable equations
                if variable.is_local():
                    v = c[variable.name()]

                    # At this point, the model has been validated so there can
                    # only be one variable without an rhs or initial value, and
                    # this is either known to be the free variable or assumed
                    # to be (if not explicitly set by the user).

                    # Add equation
                    rhs = variable.rhs_or_initial_value()
                    if variable.is_free() or rhs is None:
                        # Use zero for free or undefined variables
                        v.set_rhs(
                            myokit.Number(0, variable.units().myokit_unit()))
                        if variable.is_free():
                            v.set_binding('time')
                        else:
                            undefined_variables.append(v)
                    else:
                        # Add RHS with Myokit references
                        rhs = rhs.clone(subst=var_map)

                        v.set_rhs(rhs)
                        if variable.is_state():
                            init = variable.initial_value()
                            v.promote(0 if init is None else init)

                # Add local copies of variables requiring unit conversion
                elif variable in needs_conversion:

                    # Get Myokit variable
                    v = c[variable.name()]

                    # Get Myokit variable for source
                    r = variable.value_source()
                    r = m[r.component().name()][r.name()]

                    # Get conversion factor
                    try:
                        f = myokit.Unit.conversion_factor(r.unit(), v.unit())
                        f = myokit.Number(f)
                    except myokit.IncompatibleUnitError:
                        warnings.warn(
                            'Unable to determine unit conversion factor for '
                            + str(v) + ', from ' + str(v.unit()) + ' to '
                            + str(r.unit()) + '.')
                        f = myokit.Number(1)

                    # Add equation
                    v.set_rhs(myokit.Multiply(myokit.Name(r), f))

        # Check that a binding to time has been made
        if m.binding('time') is None:
            if undefined_variables:
                undefined_variables[0].set_binding('time')

        return m

    def name(self):
        """
        Returns this model's name.
        """
        return self._name

    def _register_cmeta_id(self, cmeta_id, element):
        """
        Should be called whenever a cmeta:id is set for a CellML element in
        this model (or for the model itself).

        Arguments

        ``cmeta_id``
            A cmeta:id string, unique within this model.
        ``element``
            An :class:`AnnotatableElement` object.

        """
        # Check
        cmeta_id = str(cmeta_id).strip()
        if not cmeta_id:
            raise CellMLError('A cmeta:id must be a non-empty string.')
        if cmeta_id in self._cmeta_ids:
            raise CellMLError(
                'A cmeta:id must be unique within a model (8.5.1).')

        # And store
        self._cmeta_ids[cmeta_id] = element

    def set_free_variable(self, variable):
        """
        Tells this model which variable to regard as the free variable (with
        respect to which derivatives are taken).
        """
        if variable is not None and variable._model is not self:
            raise ValueError('Given variable must be from this model.')

        # Check interface
        if variable is not None and (variable._public_interface == 'in' or
                                     variable._private_interface == 'in'):
            raise CellMLError(
                'The free variable cannot have an "in" interface.')

        # Unset previous
        if self._free_variable is not None:
            self._free_variable._is_free = False

        # Set new
        self._free_variable = variable
        if variable is not None:
            variable._is_free = True

    def __str__(self):
        return 'Model[@name="' + self._name + '"]'

    def units(self):
        """
        Returns an iterator over the :class:`Units` objects in this model.
        """
        # Note: Must use _by_name here, other one doesn't necessarily contain
        # all units objects (only names are unique).
        return iter(self._units.values())

    def _unregister_cmeta_id(self, cmeta_id):
        """
        Should be called whenever a cmeta:id for a CellML element in this model
        (or for this model itself) is unset.
        """
        del self._cmeta_ids[cmeta_id]

    def validate(self):
        """
        Validates this model, raising a :class:`CellMLError` if an errors are
        found.
        """
        # Validate components and variables
        for c in self._components.values():
            c._validate()

        # Check at most one variable doesn't have a source (one free variable
        # is allowed)
        free = set()
        for c in self:
            for v in c:
                if v.value_source() is v:
                    if v.rhs() is None and v.initial_value() is None:
                        free.add(v)

        if len(free) > 1:
            warnings.warn('More than one variable does not have a value.')

        elif self._free_variable is not None and len(free) == 1:
            free = free.pop()
            if self._free_variable is not free:
                warnings.warn(
                    'No value is defined for the variable "'
                    + free.name() + '", but "' + self._free_variable.name()
                    + '" is listed as the free variable.')

    def version(self):
        """
        Returns the CellML version this model is in.
        """
        return self._version


class Units:
    """
    Represents a CellML units definition, should not be created directly but
    only via :meth:`Model.add_units()` or :meth:`Component.add_units()`.

    Arguments:

    ``name``
        A string name, used to refer to these units in the model.
    ``myokit_unit``
        A :class:`myokit.Unit` representing these units.
    ``predefined``
        Set to ``True`` when creating an object for a predefined name. This is
        only used internally.

    """
    def __init__(self, name, myokit_unit, predefined=False):

        # Check and store name
        if not is_valid_identifier(name):
            raise CellMLError(
                'Units name must be a valid CellML identifier (5.4.1.2).')
        if not predefined and name in self._si_units:
            raise CellMLError(
                'Units name "' + name + '" overlaps with a predefined name'
                ' (5.4.1.2).')
        self._name = name

        # Check and store Myokit unit
        if not isinstance(myokit_unit, myokit.Unit):
            raise ValueError(
                'The argument `myokit_unit` must be a myokit.Unit object.')
        self._myokit_unit = myokit_unit

    @classmethod
    def find_units(cls, name):
        """
        Searches for a predefined unit with the given ``name`` and returns a
        :class:`Units` object representing it.

        If no such unit is found a :class:`CellMLError` is raised.
        """
        # Check for unsupported units
        if name == 'celsius':
            raise UnsupportedBaseUnitsError('celsius')

        # Check if we have a cached object for this
        obj = cls._si_unit_objects.get(name, None)
        if obj is None:

            # Check if this is an si unit
            myokit_unit = cls._si_units.get(name, None)

            # If not raise error (and one that makes sense even if this was
            # called via a model or component units lookup).
            if myokit_unit is None:
                raise CellMLError('Unknown units name "' + str(name) + '".')

            # Create and store object
            obj = cls(name, myokit_unit, predefined=True)
            cls._si_unit_objects[name] = obj

        # Return
        return obj

    @classmethod
    def find_units_name(cls, myokit_unit):
        """
        Attempts to find a string name for the given :class:`myokit.Unit`.

        Raises a :class:`CellMLError` is no appropriate unit is found.
        """
        try:
            return cls._si_units_r[myokit_unit]
        except KeyError:
            raise CellMLError(
                'No name found for myokit unit ' + str(myokit_unit) + '.')

    def myokit_unit(self):
        """
        Returns the :class:`Myokit.Unit` equivalent for this units definition.
        """
        return self._myokit_unit

    def name(self):
        """
        Returns the string name for this units definition.
        """
        return self._name

    @classmethod
    def parse_unit_row(cls, units, prefix=None, exponent=None, multiplier=None,
                       context=None):
        """
        Creates a :class:`myokit.Unit` using the information found in a single
        CellML ``unit`` element.

        Arguments

        ``units``
            The name of the units to start from.
        ``prefix``
            An optional argument that can be either a string from one of the
            predefined units prefixes, or an integer. If an integer is given
            the unit is scaled by ``10**prefix``. Used to make e.g. ``cm``.
        ``exponent``
            An optional exponent (power) for this unit, used to make e.g.
            ``m**2`` or ``cm**2``.
        ``multiplier``
            An optional multiplier for this unit, used to make e.g. ``inches``.
            Note that exponentiation affects prefixes, but not multipliers.
        ``context``
            An optional :class:`Model` or :class:`Component` to use when
            looking up ``units`` names. If not given, only the SI units will be
            supported.

        """
        # Find units object to start from
        if context is None:
            unit = cls.find_units(units)
        else:
            unit = context.find_units(units)
        unit = unit.myokit_unit()

        # Handle prefix
        if prefix is not None:
            p = cls._si_prefixes.get(prefix, None)
            if p is None:
                try:
                    # Test if can convert to float
                    p = float(prefix)
                except ValueError:
                    pass
                else:
                    # Test if is integer
                    if int(p) != p:
                        p = None
            if p is None:
                raise CellMLError(
                    'Units prefix must be a string from the list of known'
                    ' prefixes or an integer (5.4.2.3).')

            # Apply prefix to unit

            # float(10**309) is the first int that doesn't fit in a float
            if p > 309:
                raise CellMLError('Unit prefix too large: 10^' + str(p))
            unit *= 10**int(p)

        # Handle exponent (note: prefix is exponentiated, multiplier is not).
        if exponent is not None:
            try:
                e = float(exponent)
            except ValueError:
                raise CellMLError(
                    'Unit exponent must be a real number (5.4.2.4).')

            # Apply exponent to unit
            unit **= e

        # Handle multiplier
        if multiplier is not None:
            try:
                m = float(multiplier)
            except ValueError:
                raise CellMLError(
                    'Unit multiplier must be a real number (5.4.2.5).')

            # Apply multiplier to unit
            unit *= m

        # Return
        return unit

    @classmethod
    def si_unit_names(cls):
        """
        Returns an iterator over the predefined unit names.
        """
        return cls._si_units.keys()

    def __str__(self):
        return 'Units[@name="' + self._name + '"]'

    # Predefined units in CellML, name to Unit
    _si_units = {
        'ampere': myokit.units.A,
        'becquerel': myokit.units.Bq,
        'candela': myokit.units.cd,
        'celsius': None,
        'coulomb': myokit.units.C,
        'dimensionless': myokit.units.dimensionless,
        'farad': myokit.units.F,
        'gram': myokit.units.g,
        'gray': myokit.units.Gy,
        'henry': myokit.units.H,
        'hertz': myokit.units.Hz,
        'joule': myokit.units.J,
        'katal': myokit.units.kat,
        'kelvin': myokit.units.K,
        'kilogram': myokit.units.kg,
        'liter': myokit.units.L,
        'litre': myokit.units.L,
        'lumen': myokit.units.lm,
        'lux': myokit.units.lux,
        'meter': myokit.units.m,
        'metre': myokit.units.m,
        'mole': myokit.units.mol,
        'newton': myokit.units.N,
        'ohm': myokit.units.R,
        'pascal': myokit.units.Pa,
        'radian': myokit.units.radian,
        'second': myokit.units.s,
        'siemens': myokit.units.S,
        'sievert': myokit.units.Sv,
        'steradian': myokit.units.steradian,
        'tesla': myokit.units.T,
        'volt': myokit.units.V,
        'watt': myokit.units.W,
        'weber': myokit.units.Wb,
    }

    # Predefined Units objects
    _si_unit_objects = {}

    # Predefined units in CellML, Unit object to name
    _si_units_r = {
        myokit.units.A: 'ampere',
        myokit.units.Bq: 'becquerel',
        myokit.units.cd: 'candela',
        myokit.units.C: 'coulomb',
        myokit.units.dimensionless: 'dimensionless',
        myokit.units.F: 'farad',
        myokit.units.g: 'gram',
        myokit.units.Gy: 'gray',
        myokit.units.H: 'henry',
        myokit.units.hertz: 'hertz',
        myokit.units.joule: 'joule',
        myokit.units.kat: 'katal',
        myokit.units.K: 'kelvin',
        myokit.units.kg: 'kilogram',
        myokit.units.L: 'litre',
        myokit.units.lumen: 'lumen',
        myokit.units.lux: 'lux',
        myokit.units.m: 'metre',
        myokit.units.mol: 'mole',
        myokit.units.N: 'newton',
        myokit.units.ohm: 'ohm',
        myokit.units.Pa: 'pascal',
        myokit.units.s: 'second',
        myokit.units.S: 'siemens',
        myokit.units.sievert: 'sievert',
        myokit.units.T: 'tesla',
        myokit.units.V: 'volt',
        myokit.units.W: 'watt',
        myokit.units.weber: 'weber',
    }

    # Recognised unit prefixes, name to multiplier
    _si_prefixes = {
        'yotta': 24,
        'zetta': 21,
        'exa': 18,
        'peta': 15,
        'tera': 12,
        'giga': 9,
        'mega': 6,
        'kilo': 3,
        'hecto': 2,
        'deka': 1,            # US Style, as defined in CellML 1.0
        'deca': 1,            # SI prefix
        'deci': -1,
        'centi': -2,
        'milli': -3,
        'micro': -6,
        'nano': -9,
        'pico': -12,
        'femto': -15,
        'atto': -18,
        'zepto': -21,
        'yocto': -24,
    }


class Variable(AnnotatableElement):
    """
    Represents a model variable, should not be created directly but only via
    :meth:`Component.add_variable()`.

    Arguments

    ``component``
        This variable's parent :class:`Component`.
    ``name``
        This variable's name (a valid CellML identifier string).
    ``units``
        The string name of a units object known to this variable's component.
    ``public_interface``
        The variable's public interface.
    ``private_interface``
        The variable's private interface.

    """
    def __init__(self, component, name, units, public_interface='none',
                 private_interface='none'):
        super().__init__(component.model())

        # Store component
        self._component = component

        # Check and store name
        if not is_valid_identifier(name):
            raise CellMLError(
                'Variable name must be a valid CellML identifier (3.4.3.2).')
        self._name = name

        # Check and store units
        try:
            self._units = component.find_units(units)
        except UnsupportedBaseUnitsError as e:
            raise UnsupportedBaseUnitsError(
                'Variable units attribute references the unsupported base'
                ' units "' + e.units + '".')
        except CellMLError:
            raise CellMLError(
                'Variable units attribute must reference a units element in'
                ' the current component or model, or one of the predefined'
                ' units, found "' + str(units) + '" (3.4.3.3).')

        # Check and store interfaces
        if public_interface not in ['none', 'in', 'out']:
            raise CellMLError(
                'Public interface must be "in", "out", or "none" (3.4.3.4).')
        if private_interface not in ['none', 'in', 'out']:
            raise CellMLError(
                'Private interface must be "in", "out", or "none" (3.4.3.5).')
        # Can't have two "in" interfaces
        if public_interface == private_interface == 'in':
            raise CellMLError(
                'Public and private interface can not both be "in" (3.4.3.6).')
        self._public_interface = public_interface
        self._private_interface = private_interface

        # Initial value
        self._initial_value = None

        # RHS
        self._rhs = None

        # Is this a state variable?
        self._is_state = False

        # Is this the free variable?
        self._is_free = False

        # If this variable derives its value from another variable, that
        # variable is listed here.
        self._source = None

    def component(self):
        """
        Return this variable's component.
        """
        return self._component

    def initial_value(self):
        """
        Returns this variable's initial value, or ``None`` if it is not set.
        """
        return self._initial_value

    def is_local(self):
        """
        Checks if this variable defines its own value.
        """
        return self._source is None

    def is_free(self):
        """
        Checks if this variable has been marked as the free variable.

        See :meth:`Model.set_free_variable()`.
        """
        return self._is_free

    def is_state(self):
        """
        Checks if this variable has been marked as a state variable.
        """
        return self._is_state

    def model(self):
        """
        Return this variable's model.
        """
        return self._model

    def name(self):
        """
        Returns this variable's name.
        """
        return self._name

    def public_interface(self):
        """
        Returns this variable's public interface.
        """
        return self._public_interface

    def private_interface(self):
        """
        Returns this variable's private interface.
        """
        return self._private_interface

    def rhs(self):
        """
        Returns this variable's right-hand side.
        """
        return self._rhs

    def rhs_or_initial_value(self):
        """
        For non-states, returns this variable's RHS or a :class:`myokit.Number`
        representing the initial value if no RHS is set. For states always
        returns the RHS.
        """
        if self._is_state:
            return self._rhs
        if self._rhs is None and self._initial_value is not None:
            return myokit.Number(
                self._initial_value, self._units.myokit_unit())
        return self._rhs

    def set_initial_value(self, value):
        """
        Sets this variable's intial value (must be a number or ``None``).
        """
        # Allow unsetting with ``None``
        if value is None:
            self._initial_value = None
            return

        # Check interface
        if self._public_interface == 'in' or self._private_interface == 'in':
            i = 'public' if self._public_interface == 'in' else 'private'
            raise CellMLError(
                'An initial value cannot be set for ' + str(self) + ', which'
                ' has ' + i + '_interface="in" (3.4.3.8).')

        # Check and store
        try:
            self._initial_value = float(value)
        except ValueError:
            if self._model.version() == '1.0':
                raise CellMLError(
                    'If given, a variable initial_value must be a real number'
                    ' (3.4.3.7).')
            else:
                raise CellMLError(
                    'If given, a variable initial_value must be a real number'
                    ' (using variables as initial values is not supported).')

    def set_is_state(self, state):
        """
        Set whether this variable is a state variable or not.
        """
        # Check interface
        if self._public_interface == 'in' or self._private_interface == 'in':
            i = 'public' if self._public_interface == 'in' else 'private'
            raise CellMLError(
                'State variables can not have an "in" interface.')

        self._is_state = bool(state)

    def set_rhs(self, rhs):
        """
        Sets a right-hand side expression for this variable.

        The given ``rhs`` must be a :class:`myokit.Expression` tree where any
        :class:`myokit.Name` objects have a CellML :class:`Variable` as their
        value.
        """
        # Check interface
        if self._public_interface == 'in' or self._private_interface == 'in':
            i = 'public' if self._public_interface == 'in' else 'private'
            raise CellMLError(
                'An equation cannot be set for ' + str(self) + ', which has '
                + i + '_interface="in" (4.4.4).')

        # Check all references in equation are known and local
        if rhs is not None:
            for ref in rhs.references():
                var = ref.var()
                if var._component is not self._component:
                    raise CellMLError(
                        'A variable RHS can only reference variables from the'
                        ' same component, found: ' + str(var) + '.')

            # Check all units are known
            numbers_without_units = {}
            for x in rhs.walk(myokit.Number):
                # Replace None with dimensionless
                if x.unit() is None:
                    numbers_without_units[x] = myokit.Number(
                        x.eval(), myokit.units.dimensionless)
                else:
                    try:
                        self._component.find_units_name(x.unit())
                    except CellMLError:
                        raise CellMLError(
                            'All units appearing in a variable\'s RHS must be'
                            ' known to its component, found: ' + str(x.unit())
                            + '.')
            if numbers_without_units:
                rhs = rhs.clone(subst=numbers_without_units)

        # Store
        self._rhs = rhs

    def source(self):
        """
        If this :class:`Variable` has an "in" interface and is connected to
        another variable, this method returns that variable.

        If not, it returns ``None``.
        """
        return self._source

    def __str__(self):
        return (
            'Variable[@name="' + self._name + '"] in ' + str(self._component))

    def units(self):
        """
        Returns the units this variable is in (as a :class:`Units` object).
        """
        return self._units

    def _validate(self):
        """
        Validates this variable, raising a :class:`CellMLError` if any errors
        are found.
        """
        # Check that variables with an in interface are connected
        # Sort of allowed in the spec ?
        if self._public_interface == 'in' or self._private_interface == 'in':
            i = 'public' if self._public_interface == 'in' else 'private'
            if self._source is None:
                warnings.warn(
                    str(self) + ' has ' + i + '_interface="in", but is not'
                    ' connected to a variable with an appropriate "out"')

        # Check that state variables define two values
        elif self._is_state:

            if self._initial_value is None:
                warnings.warn(
                    'State ' + str(self) + ' has no initial value.')

            if self._rhs is None:
                raise CellMLError(
                    'State ' + str(self) + ' must have a defining equation.')

        # Check that other variables define a value
        elif self._rhs is None:
            if self._initial_value is None and not self._is_free:
                warnings.warn('No value set for ' + str(self) + '.')

        # And only one value
        elif self._initial_value is not None:
            raise CellMLError(
                'Overdefined: ' + str(self) + ' has both an initial value and'
                ' a defining equation (which is not an ODE).')

    def value_source(self):
        """
        Returns the :class:`Variable` that this variable derives its value
        from.

        If the variable has a defining equation or an initial value, this will
        return the variable itself. If it is connected to another variable, it
        will follow the chain of dependencies until a variable with an equation
        or value is found.
        """
        var = self
        while var._source is not None:
            var = var._source
            assert var is not self
        return var




#!/usr/bin/env python3
#
# Tests the C implementation of the Pacing System
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest
import numpy as np

import myokit

from myokit.tests.ansic_event_based_pacing import AnsicEventBasedPacing
from myokit.tests.ansic_time_series_pacing import AnsicTimeSeriesPacing


class EventBasedPacingAnsicTest(unittest.TestCase):
    """
    Tests the C implementation of event based pacing.
    """
    def test_with_event_at_t_0(self):
        # Test with event starting at t=0

        # Test basics
        p = myokit.Protocol()
        p.schedule(2, 0, 1, 10, 0)
        s = AnsicEventBasedPacing(p)
        self.assertEqual(s.time(), 0)
        self.assertEqual(s.next_time(), 1)
        self.assertEqual(s.pace(), 2)
        s.advance(0)
        self.assertEqual(s.time(), 0)
        self.assertEqual(s.next_time(), 1)
        self.assertEqual(s.pace(), 2)
        s.advance(0)
        s.advance(0)
        s.advance(0)
        self.assertEqual(s.time(), 0)
        self.assertEqual(s.next_time(), 1)
        self.assertEqual(s.pace(), 2)
        s.advance(0.5)
        self.assertEqual(s.time(), 0.5)
        self.assertEqual(s.next_time(), 1)
        self.assertEqual(s.pace(), 2)
        s.advance(1)
        self.assertEqual(s.time(), 1)
        self.assertEqual(s.next_time(), 10)
        self.assertEqual(s.pace(), 0)
        s.advance(2)
        self.assertEqual(s.time(), 2)
        self.assertEqual(s.next_time(), 10)
        self.assertEqual(s.pace(), 0)
        s.advance(10)
        self.assertEqual(s.time(), 10)
        self.assertEqual(s.next_time(), 11)
        self.assertEqual(s.pace(), 2)

    def test_with_event_at_t_1(self):
        p = myokit.Protocol()
        p.schedule(2, 1, 1, 10, 0)
        s = AnsicEventBasedPacing(p)
        self.assertEqual(s.time(), 0)
        self.assertEqual(s.next_time(), 1)
        self.assertEqual(s.pace(), 0)
        p = myokit.Protocol()
        s = AnsicEventBasedPacing(p)
        self.assertEqual(s.time(), 0)
        self.assertTrue(s.next_time() > 1e123)
        self.assertEqual(s.pace(), 0)

    def test_simultaneous_event_error(self):
        # Test raising of errors on rescheduled periodic events

        p = myokit.Protocol()
        p.schedule(1, 0, 1, 1000)
        p.schedule(1, 3000, 1)
        s = AnsicEventBasedPacing(p)
        t = s.next_time()
        self.assertEqual(t, 1)
        s.advance(t)
        t = s.next_time()
        self.assertEqual(t, 1000)
        s.advance(t)
        t = s.next_time()
        self.assertEqual(t, 1001)
        s.advance(t)
        t = s.next_time()
        self.assertEqual(t, 2000)
        self.assertRaises(myokit.SimultaneousProtocolEventError, s.advance, t)

        # Test raising of errors in simulation
        p = myokit.Protocol()
        p.schedule(1, 0, 1, 10)
        p.schedule(1, 30, 1)
        m = myokit.load_model('example')
        s = myokit.Simulation(m, p)
        self.assertRaises(myokit.SimultaneousProtocolEventError, s.run, 40)

    def test_negative_time(self):
        # Test starting from a negative time

        p = myokit.pacing.blocktrain(level=1, duration=1, period=2)
        s = AnsicEventBasedPacing(p, initial_time=-100)
        self.assertEqual(s.time(), -100)
        self.assertEqual(s.next_time(), 0)
        self.assertEqual(s.pace(), 0)

        p = myokit.pacing.blocktrain(level=1, duration=1, period=2, offset=1)
        s = AnsicEventBasedPacing(p, initial_time=-100)
        self.assertEqual(s.time(), -100)
        self.assertEqual(s.next_time(), 1)
        self.assertEqual(s.pace(), 0)
        s.advance(s.next_time())
        self.assertEqual(s.time(), 1)
        self.assertEqual(s.pace(), 1)


class TimeSeriesPacingAnsicTest(unittest.TestCase):
    """
    Test the Ansi-C time-series pacing system (aka point-list, aka data-clamp).
    """
    def test_time_series_pacing_ansic(self):
        # Tests if the basics work

        if False:
            # Graphical test, just for playing with the pacing system
            m, p, x = myokit.load('example')
            s = myokit.Simulation(m, p)
            d = s.run(500).npview()
            # Get time and voltage
            t = d.time()
            v = d['membrane.V']
            # Plot trace
            import matplotlib.pyplot as plt
            plt.figure()
            plt.plot(t, v, 'x')
            # Get some points halfway, 1/4 of the way, and 3/4 of the way
            # between the known points
            t2 = t[:-1] + 0.25 * (t[1:] - t[:-1])
            t3 = t[:-1] + 0.5 * (t[1:] - t[:-1])
            t4 = t[:-1] + 0.75 * (t[1:] - t[:-1])
            t2 = np.concatenate((t2, t3, t4))
            del t3, t4
            # Get the pacing value at the points, measure how long it takes
            pacing = AnsicFixedFormPacing(list(t), list(v))
            b = myokit.tools.Benchmarker()
            v2 = [pacing.pace(x) for x in t2]
            print(b.time())
            # Plot the new points
            plt.plot(t2, v2, '.', color='green')
            plt.show()
            # Quite
            import sys
            sys.exit(1)

        # Test with small lists
        times = [0, 0, 1, 1, 1, 2, 2, 2, 3, 4, 5, 7]
        values = list(range(len(times)))
        pacing = AnsicTimeSeriesPacing(
            myokit.TimeSeriesProtocol(times, values))

        self.assertEqual(pacing.pace(-1), 0)
        self.assertEqual(pacing.pace(0), 1)
        self.assertEqual(pacing.pace(1), 2)
        self.assertEqual(pacing.pace(2), 5)
        self.assertEqual(pacing.pace(3), 8)
        self.assertEqual(pacing.pace(4), 9)
        self.assertEqual(pacing.pace(5), 10)
        self.assertEqual(pacing.pace(7), 11)
        self.assertEqual(pacing.pace(8), 11)
        self.assertEqual(pacing.pace(1.5), 4.5)
        self.assertEqual(pacing.pace(1.75), 4.75)
        self.assertEqual(pacing.pace(6), 10.5)
        self.assertEqual(pacing.pace(5.5), 10.25)

        # Test not starting at 0
        times = [1, 2]
        values = [10, 20]
        pacing = AnsicTimeSeriesPacing(
            myokit.TimeSeriesProtocol(times, values))

        self.assertEqual(pacing.pace(-1), 10)
        self.assertEqual(pacing.pace(0), 10)
        self.assertEqual(pacing.pace(1), 10)
        self.assertEqual(pacing.pace(2), 20)
        self.assertEqual(pacing.pace(3), 20)
        self.assertEqual(pacing.pace(1.5), 15)


if __name__ == '__main__':
    unittest.main()



#
# Stan expression writer and keywords
#
# cell.stan:: This will become the stan model definition file
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit

from myokit.formats.python import PythonExpressionWriter


class StanExpressionWriter(PythonExpressionWriter):
    """
    This:class:`ExpressionWriter <myokit.formats.ExpressionWriter>` translates
    Myokit:class:`expressions <myokit.Expression>` to a Stan syntax.
    """
    def __init__(self):
        super().__init__()
        self._function_prefix = ''

        self._fcond = None
        self.set_condition_function('ifthenelse')

    def set_condition_function(self, func=None):
        """
        Sets a function name to use for if statements

        By setting func to None you can revert back to the default behavior
         (the ternary operator). Any other value will be interpreted as the
         name of a function taking arguments (condition, value_if_true,
         value_if_false).
        """
        self._fcond = func

    #def _ex_name(self, e):
    #def _ex_derivative(self, e):
    #def _ex_number(self, e):
    #def _ex_prefix_plus(self, e):
    #def _ex_prefix_minus(self, e):
    #def _ex_plus(self, e):
    #def _ex_minus(self, e):
    #def _ex_multiply(self, e):
    #def _ex_divide(self, e):

    def _ex_quotient(self, e):
        # Use floor to round towards minus infinity
        return self.ex(myokit.Floor(myokit.Divide(e[0], e[1])))

    def _ex_remainder(self, e):
        # fmod uses correct convention
        return self._ex_function(e, 'fmod')

    def _ex_power(self, e):
        return self._ex_infix(e, '^')

    #def _ex_sqrt(self, e):
    #def _ex_sin(self, e):
    #def _ex_cos(self, e):
    #def _ex_tan(self, e):
    #def _ex_asin(self, e):
    #def _ex_acos(self, e):
    #def _ex_atan(self, e):
    #def _ex_exp(self, e):

    def _ex_log(self, e):
        if len(e) == 1:
            return self._ex_function(e, 'log')
        else:
            return '(log(' + self.ex(e[0]) + ') / log(' + self.ex(e[1]) + '))'

    def _ex_log10(self, e):
        return 'log10(' + self.ex(e[0]) + ')'

    #def _ex_floor(self, e):
    #def _ex_ceil(self, e):

    def _ex_abs(self, e):
        return self._ex_function(e, 'abs')

    def _ex_not(self, e):
        return '!(' + self.ex(e[0]) + ')'

    #def _ex_equal(self, e):
    #def _ex_not_equal(self, e):
    #def _ex_more(self, e):
    #def _ex_less(self, e):
    #def _ex_more_equal(self, e):
    #def _ex_less_equal(self, e):

    def _ex_and(self, e):
        return self._ex_infix_condition(e, '&&')

    def _ex_or(self, e):
        return self._ex_infix_condition(e, '||')

    def _ex_if(self, e):
        return ('(' + self.ex(e._i) + ' ? ' + self.ex(e._t) + ' : ' +
                self.ex(e._e) + ')')

    def _ex_piecewise(self, e):
        s = []
        n = len(e._i)
        for i in range(0, n):
            s.append('(%s ? %s : ' % (self.ex(e._i[i]), self.ex(e._e[i])))
        s.append(self.ex(e._e[n]))
        s.append(')' * n)
        return ''.join(s)




#!/usr/bin/env python3
#
# Tests the expression writer classes.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import myokit
import myokit.formats
import myokit.formats.ansic
import myokit.formats.cpp
import myokit.formats.latex
import myokit.formats.matlab
import myokit.formats.python
import myokit.formats.stan


class ExpressionWriterTest(unittest.TestCase):
    """ Test shared expression writer functionality. """

    def test_ewriter_interface(self):
        # Test listing and creating expression writers.

        # Test listing
        es = myokit.formats.ewriters()
        self.assertTrue(len(es) > 0)

        # Create one of each
        for e in es:
            self.assertIsInstance(e, str)
            e = myokit.formats.ewriter(e)
            self.assertTrue(isinstance(e, myokit.formats.ExpressionWriter))

    def test_unknown(self):
        # Test requesting an unknown expression writer.
        # Test fetching using ewriter method
        self.assertRaisesRegex(
            KeyError, 'Expression writer not found', myokit.formats.ewriter,
            'dada')


class AnsicExpressionWriterTest(unittest.TestCase):
    """ Test the Ansi C ewriter class. """

    def test_all(self):
        w = myokit.formats.ansic.AnsiCExpressionWriter()

        model = myokit.Model()
        component = model.add_component('c')
        avar = component.add_variable('a')

        # Name
        a = myokit.Name(avar)
        self.assertEqual(w.ex(a), 'c.a')
        # Number with unit
        b = myokit.Number('12', 'pF')
        self.assertEqual(w.ex(b), '12.0')

        # Prefix plus
        x = myokit.PrefixPlus(b)
        self.assertEqual(w.ex(x), '12.0')
        # Prefix minus
        x = myokit.PrefixMinus(b)
        self.assertEqual(w.ex(x), '(-12.0)')

        # Plus
        x = myokit.Plus(a, b)
        self.assertEqual(w.ex(x), 'c.a + 12.0')
        # Minus
        x = myokit.Minus(a, b)
        self.assertEqual(w.ex(x), 'c.a - 12.0')
        # Multiply
        x = myokit.Multiply(a, b)
        self.assertEqual(w.ex(x), 'c.a * 12.0')
        # Divide
        x = myokit.Divide(a, b)
        self.assertEqual(w.ex(x), 'c.a / 12.0')

        # Quotient
        x = myokit.Quotient(a, b)
        self.assertEqual(w.ex(x), 'floor(c.a / 12.0)')
        # Remainder
        x = myokit.Remainder(a, b)
        self.assertEqual(w.ex(x), 'c.a - 12.0 * (floor(c.a / 12.0))')

        # Power
        x = myokit.Power(a, b)
        self.assertEqual(w.ex(x), 'pow(c.a, 12.0)')
        # Sqrt
        x = myokit.Sqrt(b)
        self.assertEqual(w.ex(x), 'sqrt(12.0)')
        # Exp
        x = myokit.Exp(a)
        self.assertEqual(w.ex(x), 'exp(c.a)')
        # Log(a)
        x = myokit.Log(b)
        self.assertEqual(w.ex(x), 'log(12.0)')
        # Log(a, b)
        x = myokit.Log(a, b)
        self.assertEqual(w.ex(x), '(log(c.a) / log(12.0))')
        # Log10
        x = myokit.Log10(b)
        self.assertEqual(w.ex(x), 'log10(12.0)')

        # Sin
        x = myokit.Sin(b)
        self.assertEqual(w.ex(x), 'sin(12.0)')
        # Cos
        x = myokit.Cos(b)
        self.assertEqual(w.ex(x), 'cos(12.0)')
        # Tan
        x = myokit.Tan(b)
        self.assertEqual(w.ex(x), 'tan(12.0)')
        # ASin
        x = myokit.ASin(b)
        self.assertEqual(w.ex(x), 'asin(12.0)')
        # ACos
        x = myokit.ACos(b)
        self.assertEqual(w.ex(x), 'acos(12.0)')
        # ATan
        x = myokit.ATan(b)
        self.assertEqual(w.ex(x), 'atan(12.0)')

        # Floor
        x = myokit.Floor(b)
        self.assertEqual(w.ex(x), 'floor(12.0)')
        # Ceil
        x = myokit.Ceil(b)
        self.assertEqual(w.ex(x), 'ceil(12.0)')
        # Abs
        x = myokit.Abs(b)
        self.assertEqual(w.ex(x), 'fabs(12.0)')

        # Equal
        x = myokit.Equal(a, b)
        self.assertEqual(w.ex(x), '(c.a == 12.0)')
        # NotEqual
        x = myokit.NotEqual(a, b)
        self.assertEqual(w.ex(x), '(c.a != 12.0)')
        # More
        x = myokit.More(a, b)
        self.assertEqual(w.ex(x), '(c.a > 12.0)')
        # Less
        x = myokit.Less(a, b)
        self.assertEqual(w.ex(x), '(c.a < 12.0)')
        # MoreEqual
        x = myokit.MoreEqual(a, b)
        self.assertEqual(w.ex(x), '(c.a >= 12.0)')
        # LessEqual
        x = myokit.LessEqual(a, b)
        self.assertEqual(w.ex(x), '(c.a <= 12.0)')

        # Not
        cond1 = myokit.parse_expression('5 > 3')
        cond2 = myokit.parse_expression('2 < 1')
        x = myokit.Not(cond1)
        self.assertEqual(w.ex(x), '!((5.0 > 3.0))')
        # And
        x = myokit.And(cond1, cond2)
        self.assertEqual(w.ex(x), '((5.0 > 3.0) && (2.0 < 1.0))')
        # Or
        x = myokit.Or(cond1, cond2)
        self.assertEqual(w.ex(x), '((5.0 > 3.0) || (2.0 < 1.0))')

        # If
        x = myokit.If(cond1, a, b)
        self.assertEqual(w.ex(x), '((5.0 > 3.0) ? c.a : 12.0)')
        # Piecewise
        c = myokit.Number(1)
        x = myokit.Piecewise(cond1, a, cond2, b, c)
        self.assertEqual(
            w.ex(x),
            '((5.0 > 3.0) ? c.a : ((2.0 < 1.0) ? 12.0 : 1.0))')

        # If/Piecewise with special function
        w.set_condition_function('ifthenelse')
        x = myokit.If(cond1, a, b)
        self.assertEqual(w.ex(x), 'ifthenelse((5.0 > 3.0), c.a, 12.0)')
        # Piecewise
        c = myokit.Number(1)
        x = myokit.Piecewise(cond1, a, cond2, b, c)
        self.assertEqual(
            w.ex(x),
            'ifthenelse((5.0 > 3.0), c.a, ifthenelse((2.0 < 1.0), 12.0, 1.0))')

        # Test fetching using ewriter method
        w = myokit.formats.ewriter('ansic')
        self.assertIsInstance(w, myokit.formats.ansic.AnsiCExpressionWriter)

        # Test without a Myokit expression
        self.assertRaisesRegex(
            ValueError, 'Unknown expression type', w.ex, 7)


class CppExpressionWriterTest(unittest.TestCase):
    """ Test the C++ ewriter class. """

    def test_all(self):
        w = myokit.formats.cpp.CppExpressionWriter()

        model = myokit.Model()
        component = model.add_component('c')
        avar = component.add_variable('a')

        # Name
        a = myokit.Name(avar)
        self.assertEqual(w.ex(a), 'c.a')
        # Number with unit
        b = myokit.Number('12', 'pF')
        self.assertEqual(w.ex(b), '12.0')

        # Prefix plus
        x = myokit.PrefixPlus(b)
        self.assertEqual(w.ex(x), '12.0')
        # Prefix minus
        x = myokit.PrefixMinus(b)
        self.assertEqual(w.ex(x), '(-12.0)')

        # Plus
        x = myokit.Plus(a, b)
        self.assertEqual(w.ex(x), 'c.a + 12.0')
        # Minus
        x = myokit.Minus(a, b)
        self.assertEqual(w.ex(x), 'c.a - 12.0')
        # Multiply
        x = myokit.Multiply(a, b)
        self.assertEqual(w.ex(x), 'c.a * 12.0')
        # Divide
        x = myokit.Divide(a, b)
        self.assertEqual(w.ex(x), 'c.a / 12.0')

        # Quotient
        x = myokit.Quotient(a, b)
        self.assertEqual(w.ex(x), 'floor(c.a / 12.0)')
        # Remainder
        x = myokit.Remainder(a, b)
        self.assertEqual(w.ex(x), 'c.a - 12.0 * (floor(c.a / 12.0))')

        # Power
        x = myokit.Power(a, b)
        self.assertEqual(w.ex(x), 'pow(c.a, 12.0)')
        # Sqrt
        x = myokit.Sqrt(b)
        self.assertEqual(w.ex(x), 'sqrt(12.0)')
        # Exp
        x = myokit.Exp(a)
        self.assertEqual(w.ex(x), 'exp(c.a)')
        # Log(a)
        x = myokit.Log(b)
        self.assertEqual(w.ex(x), 'log(12.0)')
        # Log(a, b)
        x = myokit.Log(a, b)
        self.assertEqual(w.ex(x), '(log(c.a) / log(12.0))')
        # Log10
        x = myokit.Log10(b)
        self.assertEqual(w.ex(x), 'log10(12.0)')

        # Sin
        x = myokit.Sin(b)
        self.assertEqual(w.ex(x), 'sin(12.0)')
        # Cos
        x = myokit.Cos(b)
        self.assertEqual(w.ex(x), 'cos(12.0)')
        # Tan
        x = myokit.Tan(b)
        self.assertEqual(w.ex(x), 'tan(12.0)')
        # ASin
        x = myokit.ASin(b)
        self.assertEqual(w.ex(x), 'asin(12.0)')
        # ACos
        x = myokit.ACos(b)
        self.assertEqual(w.ex(x), 'acos(12.0)')
        # ATan
        x = myokit.ATan(b)
        self.assertEqual(w.ex(x), 'atan(12.0)')

        # Floor
        x = myokit.Floor(b)
        self.assertEqual(w.ex(x), 'floor(12.0)')
        # Ceil
        x = myokit.Ceil(b)
        self.assertEqual(w.ex(x), 'ceil(12.0)')
        # Abs
        x = myokit.Abs(b)
        self.assertEqual(w.ex(x), 'fabs(12.0)')

        # Equal
        x = myokit.Equal(a, b)
        self.assertEqual(w.ex(x), '(c.a == 12.0)')
        # NotEqual
        x = myokit.NotEqual(a, b)
        self.assertEqual(w.ex(x), '(c.a != 12.0)')
        # More
        x = myokit.More(a, b)
        self.assertEqual(w.ex(x), '(c.a > 12.0)')
        # Less
        x = myokit.Less(a, b)
        self.assertEqual(w.ex(x), '(c.a < 12.0)')
        # MoreEqual
        x = myokit.MoreEqual(a, b)
        self.assertEqual(w.ex(x), '(c.a >= 12.0)')
        # LessEqual
        x = myokit.LessEqual(a, b)
        self.assertEqual(w.ex(x), '(c.a <= 12.0)')

        # Not
        cond1 = myokit.parse_expression('5 > 3')
        cond2 = myokit.parse_expression('2 < 1')
        x = myokit.Not(cond1)
        self.assertEqual(w.ex(x), '!((5.0 > 3.0))')
        # And
        x = myokit.And(cond1, cond2)
        self.assertEqual(w.ex(x), '((5.0 > 3.0) && (2.0 < 1.0))')
        # Or
        x = myokit.Or(cond1, cond2)
        self.assertEqual(w.ex(x), '((5.0 > 3.0) || (2.0 < 1.0))')

        # If
        x = myokit.If(cond1, a, b)
        self.assertEqual(w.ex(x), '((5.0 > 3.0) ? c.a : 12.0)')
        # Piecewise
        c = myokit.Number(1)
        x = myokit.Piecewise(cond1, a, cond2, b, c)
        self.assertEqual(
            w.ex(x),
            '((5.0 > 3.0) ? c.a : ((2.0 < 1.0) ? 12.0 : 1.0))')

        # Test fetching using ewriter method
        w = myokit.formats.ewriter('cpp')
        self.assertIsInstance(w, myokit.formats.cpp.CppExpressionWriter)

        # Test without a Myokit expression
        self.assertRaisesRegex(
            ValueError, 'Unknown expression type', w.ex, 7)


class CudaExpressionWriterTest(unittest.TestCase):
    """ Test the CUDA ewriter class. """

    def test_all(self):
        # Single and double precision
        ws = myokit.formats.cuda.CudaExpressionWriter()
        wd = myokit.formats.cuda.CudaExpressionWriter(myokit.DOUBLE_PRECISION)

        model = myokit.Model()
        component = model.add_component('c')
        avar = component.add_variable('a')

        # Name
        a = myokit.Name(avar)
        self.assertEqual(ws.ex(a), 'c.a')
        self.assertEqual(wd.ex(a), 'c.a')
        # Number with unit
        b = myokit.Number('12', 'pF')
        self.assertEqual(ws.ex(b), '12.0f')
        self.assertEqual(wd.ex(b), '12.0')

        # Prefix plus
        x = myokit.PrefixPlus(b)
        self.assertEqual(ws.ex(x), '12.0f')
        self.assertEqual(wd.ex(x), '12.0')
        # Prefix minus
        x = myokit.PrefixMinus(b)
        self.assertEqual(ws.ex(x), '(-12.0f)')
        self.assertEqual(wd.ex(x), '(-12.0)')

        # Plus
        x = myokit.Plus(a, b)
        self.assertEqual(ws.ex(x), 'c.a + 12.0f')
        self.assertEqual(wd.ex(x), 'c.a + 12.0')
        # Minus
        x = myokit.Minus(a, b)
        self.assertEqual(ws.ex(x), 'c.a - 12.0f')
        self.assertEqual(wd.ex(x), 'c.a - 12.0')
        # Multiply
        x = myokit.Multiply(a, b)
        self.assertEqual(ws.ex(x), 'c.a * 12.0f')
        self.assertEqual(wd.ex(x), 'c.a * 12.0')
        # Divide
        x = myokit.Divide(a, b)
        self.assertEqual(ws.ex(x), 'c.a / 12.0f')
        self.assertEqual(wd.ex(x), 'c.a / 12.0')

        # Quotient
        x = myokit.Quotient(a, b)
        self.assertEqual(ws.ex(x), 'floorf(c.a / 12.0f)')
        self.assertEqual(wd.ex(x), 'floor(c.a / 12.0)')
        # Remainder
        x = myokit.Remainder(a, b)
        self.assertEqual(ws.ex(x), 'c.a - 12.0f * (floorf(c.a / 12.0f))')
        self.assertEqual(wd.ex(x), 'c.a - 12.0 * (floor(c.a / 12.0))')

        # Power
        x = myokit.Power(a, b)
        self.assertEqual(ws.ex(x), 'powf(c.a, 12.0f)')
        self.assertEqual(wd.ex(x), 'pow(c.a, 12.0)')
        # Square
        x = myokit.Power(a, myokit.Number(2))
        self.assertEqual(ws.ex(x), '(c.a * c.a)')
        self.assertEqual(wd.ex(x), '(c.a * c.a)')
        # Square with brackets
        x = myokit.Power(myokit.Plus(a, b), myokit.Number(2))
        self.assertEqual(ws.ex(x), '((c.a + 12.0f) * (c.a + 12.0f))')
        self.assertEqual(wd.ex(x), '((c.a + 12.0) * (c.a + 12.0))')
        # Sqrt
        x = myokit.Sqrt(b)
        self.assertEqual(ws.ex(x), 'sqrtf(12.0f)')
        self.assertEqual(wd.ex(x), 'sqrt(12.0)')
        # Exp
        x = myokit.Exp(a)
        self.assertEqual(ws.ex(x), 'expf(c.a)')
        self.assertEqual(wd.ex(x), 'exp(c.a)')
        # Log(a)
        x = myokit.Log(b)
        self.assertEqual(ws.ex(x), 'logf(12.0f)')
        self.assertEqual(wd.ex(x), 'log(12.0)')
        # Log(a, b)
        x = myokit.Log(a, b)
        self.assertEqual(ws.ex(x), '(logf(c.a) / logf(12.0f))')
        self.assertEqual(wd.ex(x), '(log(c.a) / log(12.0))')
        # Log10
        x = myokit.Log10(b)
        self.assertEqual(ws.ex(x), 'log10f(12.0f)')
        self.assertEqual(wd.ex(x), 'log10(12.0)')

        # Sin
        x = myokit.Sin(b)
        self.assertEqual(ws.ex(x), 'sinf(12.0f)')
        self.assertEqual(wd.ex(x), 'sin(12.0)')
        # Cos
        x = myokit.Cos(b)
        self.assertEqual(ws.ex(x), 'cosf(12.0f)')
        self.assertEqual(wd.ex(x), 'cos(12.0)')
        # Tan
        x = myokit.Tan(b)
        self.assertEqual(ws.ex(x), 'tanf(12.0f)')
        self.assertEqual(wd.ex(x), 'tan(12.0)')
        # ASin
        x = myokit.ASin(b)
        self.assertEqual(ws.ex(x), 'asinf(12.0f)')
        self.assertEqual(wd.ex(x), 'asin(12.0)')
        # ACos
        x = myokit.ACos(b)
        self.assertEqual(ws.ex(x), 'acosf(12.0f)')
        self.assertEqual(wd.ex(x), 'acos(12.0)')
        # ATan
        x = myokit.ATan(b)
        self.assertEqual(ws.ex(x), 'atanf(12.0f)')
        self.assertEqual(wd.ex(x), 'atan(12.0)')

        # Floor
        x = myokit.Floor(b)
        self.assertEqual(ws.ex(x), 'floorf(12.0f)')
        self.assertEqual(wd.ex(x), 'floor(12.0)')
        # Ceil
        x = myokit.Ceil(b)
        self.assertEqual(ws.ex(x), 'ceilf(12.0f)')
        self.assertEqual(wd.ex(x), 'ceil(12.0)')
        # Abs
        x = myokit.Abs(b)
        self.assertEqual(ws.ex(x), 'fabsf(12.0f)')
        self.assertEqual(wd.ex(x), 'fabs(12.0)')

        # Equal
        x = myokit.Equal(a, b)
        self.assertEqual(ws.ex(x), '(c.a == 12.0f)')
        self.assertEqual(wd.ex(x), '(c.a == 12.0)')
        # NotEqual
        x = myokit.NotEqual(a, b)
        self.assertEqual(ws.ex(x), '(c.a != 12.0f)')
        self.assertEqual(wd.ex(x), '(c.a != 12.0)')
        # More
        x = myokit.More(a, b)
        self.assertEqual(ws.ex(x), '(c.a > 12.0f)')
        self.assertEqual(wd.ex(x), '(c.a > 12.0)')
        # Less
        x = myokit.Less(a, b)
        self.assertEqual(ws.ex(x), '(c.a < 12.0f)')
        self.assertEqual(wd.ex(x), '(c.a < 12.0)')
        # MoreEqual
        x = myokit.MoreEqual(a, b)
        self.assertEqual(ws.ex(x), '(c.a >= 12.0f)')
        self.assertEqual(wd.ex(x), '(c.a >= 12.0)')
        # LessEqual
        x = myokit.LessEqual(a, b)
        self.assertEqual(ws.ex(x), '(c.a <= 12.0f)')
        self.assertEqual(wd.ex(x), '(c.a <= 12.0)')

        # Not
        cond1 = myokit.parse_expression('5 > 3')
        cond2 = myokit.parse_expression('2 < 1')
        x = myokit.Not(cond1)
        self.assertEqual(ws.ex(x), '!((5.0f > 3.0f))')
        self.assertEqual(wd.ex(x), '!((5.0 > 3.0))')
        # And
        x = myokit.And(cond1, cond2)
        self.assertEqual(ws.ex(x), '((5.0f > 3.0f) && (2.0f < 1.0f))')
        self.assertEqual(wd.ex(x), '((5.0 > 3.0) && (2.0 < 1.0))')
        # Or
        x = myokit.Or(cond1, cond2)
        self.assertEqual(ws.ex(x), '((5.0f > 3.0f) || (2.0f < 1.0f))')
        self.assertEqual(wd.ex(x), '((5.0 > 3.0) || (2.0 < 1.0))')

        # If
        x = myokit.If(cond1, a, b)
        self.assertEqual(ws.ex(x), '((5.0f > 3.0f) ? c.a : 12.0f)')
        self.assertEqual(wd.ex(x), '((5.0 > 3.0) ? c.a : 12.0)')
        # Piecewise
        c = myokit.Number(1)
        x = myokit.Piecewise(cond1, a, cond2, b, c)
        self.assertEqual(
            ws.ex(x),
            '((5.0f > 3.0f) ? c.a : ((2.0f < 1.0f) ? 12.0f : 1.0f))')
        self.assertEqual(
            wd.ex(x),
            '((5.0 > 3.0) ? c.a : ((2.0 < 1.0) ? 12.0 : 1.0))')

        # Test fetching using ewriter method
        w = myokit.formats.ewriter('cuda')
        self.assertIsInstance(w, myokit.formats.cuda.CudaExpressionWriter)

        # Test without a Myokit expression
        self.assertRaisesRegex(
            ValueError, 'Unknown expression type', w.ex, 7)


class LatexExpressionWriterTest(unittest.TestCase):
    """ Test the Latex ewriter class. """

    def test_all(self):
        w = myokit.formats.latex.LatexExpressionWriter()

        model = myokit.Model()
        component = model.add_component('c')
        avar = component.add_variable('a')

        # Model needs to be validated --> sets unames
        avar.set_rhs(12)
        avar.set_binding('time')
        model.validate()

        # Name
        a = myokit.Name(avar)
        self.assertEqual(w.ex(a), '\\text{a}')
        # Number with unit
        b = myokit.Number('12', 'pF')
        self.assertEqual(w.ex(b), '12.0')

        # Prefix plus
        x = myokit.PrefixPlus(b)
        self.assertEqual(w.ex(x), '12.0')
        # Prefix minus
        x = myokit.PrefixMinus(b)
        self.assertEqual(w.ex(x), '\\left(-12.0\\right)')
        # Prefix minus with bracket
        x = myokit.PrefixMinus(myokit.Plus(a, b))
        self.assertEqual(
            w.ex(x), '\\left(-\\left(\\text{a}+12.0\\right)\\right)')

        # Plus
        x = myokit.Plus(a, b)
        self.assertEqual(w.ex(x), '\\text{a}+12.0')
        # Minus
        x = myokit.Minus(a, b)
        self.assertEqual(w.ex(x), '\\text{a}-12.0')
        # Multiply
        x = myokit.Multiply(a, b)
        self.assertEqual(w.ex(x), '\\text{a}*12.0')
        # Divide
        x = myokit.Divide(a, b)
        self.assertEqual(w.ex(x), '\\frac{\\text{a}}{12.0}')

        # Quotient
        # Not supported in latex!
        x = myokit.Quotient(a, b)
        self.assertEqual(
            w.ex(x), '\\left\\lfloor\\frac{\\text{a}}{12.0}\\right\\rfloor')
        # Remainder
        x = myokit.Remainder(a, b)
        self.assertEqual(w.ex(x), '\\bmod\\left(\\text{a},12.0\\right)')

        # Power
        x = myokit.Power(a, b)
        self.assertEqual(w.ex(x), '\\text{a}^{12.0}')
        # Power with brackets
        x = myokit.Power(myokit.Plus(a, b), b)
        self.assertEqual(w.ex(x), '\\left(\\text{a}+12.0\\right)^{12.0}')
        # Sqrt
        x = myokit.Sqrt(b)
        self.assertEqual(w.ex(x), '\\sqrt{12.0}')
        # Exp
        x = myokit.Exp(a)
        self.assertEqual(w.ex(x), '\\exp\\left(\\text{a}\\right)')
        # Log(a)
        x = myokit.Log(b)
        self.assertEqual(w.ex(x), '\\log\\left(12.0\\right)')
        # Log(a, b)
        x = myokit.Log(a, b)
        self.assertEqual(w.ex(x), '\\log_{12.0}\\left(\\text{a}\\right)')
        # Log10
        x = myokit.Log10(b)
        self.assertEqual(w.ex(x), '\\log_{10.0}\\left(12.0\\right)')

        # Sin
        x = myokit.Sin(b)
        self.assertEqual(w.ex(x), '\\sin\\left(12.0\\right)')
        # Cos
        x = myokit.Cos(b)
        self.assertEqual(w.ex(x), '\\cos\\left(12.0\\right)')
        # Tan
        x = myokit.Tan(b)
        self.assertEqual(w.ex(x), '\\tan\\left(12.0\\right)')
        # ASin
        x = myokit.ASin(b)
        self.assertEqual(w.ex(x), '\\arcsin\\left(12.0\\right)')
        # ACos
        x = myokit.ACos(b)
        self.assertEqual(w.ex(x), '\\arccos\\left(12.0\\right)')
        # ATan
        x = myokit.ATan(b)
        self.assertEqual(w.ex(x), '\\arctan\\left(12.0\\right)')

        # Floor
        x = myokit.Floor(b)
        self.assertEqual(w.ex(x), '\\left\\lfloor{12.0}\\right\\rfloor')
        # Ceil
        x = myokit.Ceil(b)
        self.assertEqual(w.ex(x), '\\left\\lceil{12.0}\\right\\rceil')
        # Abs
        x = myokit.Abs(b)
        self.assertEqual(w.ex(x), '\\lvert{12.0}\\rvert')

        # Equal
        x = myokit.Equal(a, b)
        self.assertEqual(w.ex(x), '\\left(\\text{a}=12.0\\right)')
        # NotEqual
        x = myokit.NotEqual(a, b)
        self.assertEqual(w.ex(x), '\\left(\\text{a}\\neq12.0\\right)')
        # More
        x = myokit.More(a, b)
        self.assertEqual(w.ex(x), '\\left(\\text{a}>12.0\\right)')
        # Less
        x = myokit.Less(a, b)
        self.assertEqual(w.ex(x), '\\left(\\text{a}<12.0\\right)')
        # MoreEqual
        x = myokit.MoreEqual(a, b)
        self.assertEqual(w.ex(x), '\\left(\\text{a}\\geq12.0\\right)')
        # LessEqual
        x = myokit.LessEqual(a, b)
        self.assertEqual(w.ex(x), '\\left(\\text{a}\\leq12.0\\right)')

        # Not
        cond1 = myokit.parse_expression('5 > 3')
        cond2 = myokit.parse_expression('2 < 1')
        x = myokit.Not(cond1)
        self.assertEqual(
            w.ex(x), '\\not\\left(\\left(5.0>3.0\\right)\\right)')
        # And
        x = myokit.And(cond1, cond2)
        self.assertEqual(
            w.ex(x),
            '\\left(\\left(5.0>3.0\\right)\\and'
            '\\left(2.0<1.0\\right)\\right)')
        # Or
        x = myokit.Or(cond1, cond2)
        self.assertEqual(
            w.ex(x),
            '\\left(\\left(5.0>3.0\\right)\\or'
            '\\left(2.0<1.0\\right)\\right)')
        # If
        x = myokit.If(cond1, a, b)
        self.assertEqual(
            w.ex(x), 'if\\left(\\left(5.0>3.0\\right),\\text{a},12.0\\right)')
        # Piecewise
        c = myokit.Number(1)
        x = myokit.Piecewise(cond1, a, cond2, b, c)
        self.assertEqual(
            w.ex(x),
            'piecewise\\left(\\left(5.0>3.0\\right),\\text{a},'
            '\\left(2.0<1.0\\right),12.0,1.0\\right)')

        # Test fetching using ewriter method
        w = myokit.formats.ewriter('latex')
        self.assertIsInstance(w, myokit.formats.latex.LatexExpressionWriter)

        # Test without a Myokit expression
        self.assertRaisesRegex(
            ValueError, 'Unknown expression type', w.ex, 7)


class MatlabExpressionWriterTest(unittest.TestCase):
    """ Test the Matlab ewriter class. """

    def test_all(self):
        w = myokit.formats.matlab.MatlabExpressionWriter()

        model = myokit.Model()
        component = model.add_component('c')
        avar = component.add_variable('a')

        # Name
        a = myokit.Name(avar)
        self.assertEqual(w.ex(a), 'c.a')
        # Number with unit
        b = myokit.Number('12', 'pF')
        self.assertEqual(w.ex(b), '12.0')

        # Prefix plus
        x = myokit.PrefixPlus(b)
        self.assertEqual(w.ex(x), '12.0')
        # Prefix minus
        x = myokit.PrefixMinus(b)
        self.assertEqual(w.ex(x), '(-12.0)')

        # Plus
        x = myokit.Plus(a, b)
        self.assertEqual(w.ex(x), 'c.a + 12.0')
        # Minus
        x = myokit.Minus(a, b)
        self.assertEqual(w.ex(x), 'c.a - 12.0')
        # Multiply
        x = myokit.Multiply(a, b)
        self.assertEqual(w.ex(x), 'c.a * 12.0')
        # Divide
        x = myokit.Divide(a, b)
        self.assertEqual(w.ex(x), 'c.a / 12.0')

        # Quotient

        x = myokit.Quotient(a, b)
        self.assertEqual(w.ex(x), 'floor(c.a / 12.0)')
        # Remainder
        x = myokit.Remainder(a, b)
        self.assertEqual(w.ex(x), 'mod(c.a, 12.0)')

        # Power
        x = myokit.Power(a, b)
        self.assertEqual(w.ex(x), 'c.a ^ 12.0')
        # Sqrt
        x = myokit.Sqrt(b)
        self.assertEqual(w.ex(x), 'sqrt(12.0)')
        # Exp
        x = myokit.Exp(a)
        self.assertEqual(w.ex(x), 'exp(c.a)')
        # Log(a)
        x = myokit.Log(b)
        self.assertEqual(w.ex(x), 'log(12.0)')
        # Log(a, b)
        x = myokit.Log(a, b)
        self.assertEqual(w.ex(x), '(log(c.a) / log(12.0))')
        # Log10
        x = myokit.Log10(b)
        self.assertEqual(w.ex(x), 'log10(12.0)')

        # Sin
        x = myokit.Sin(b)
        self.assertEqual(w.ex(x), 'sin(12.0)')
        # Cos
        x = myokit.Cos(b)
        self.assertEqual(w.ex(x), 'cos(12.0)')
        # Tan
        x = myokit.Tan(b)
        self.assertEqual(w.ex(x), 'tan(12.0)')
        # ASin
        x = myokit.ASin(b)
        self.assertEqual(w.ex(x), 'asin(12.0)')
        # ACos
        x = myokit.ACos(b)
        self.assertEqual(w.ex(x), 'acos(12.0)')
        # ATan
        x = myokit.ATan(b)
        self.assertEqual(w.ex(x), 'atan(12.0)')

        # Floor
        x = myokit.Floor(b)
        self.assertEqual(w.ex(x), 'floor(12.0)')
        # Ceil
        x = myokit.Ceil(b)
        self.assertEqual(w.ex(x), 'ceil(12.0)')
        # Abs
        x = myokit.Abs(b)
        self.assertEqual(w.ex(x), 'abs(12.0)')

        # Equal
        x = myokit.Equal(a, b)
        self.assertEqual(w.ex(x), '(c.a == 12.0)')
        # NotEqual
        x = myokit.NotEqual(a, b)
        self.assertEqual(w.ex(x), '(c.a != 12.0)')
        # More
        x = myokit.More(a, b)
        self.assertEqual(w.ex(x), '(c.a > 12.0)')
        # Less
        x = myokit.Less(a, b)
        self.assertEqual(w.ex(x), '(c.a < 12.0)')
        # MoreEqual
        x = myokit.MoreEqual(a, b)
        self.assertEqual(w.ex(x), '(c.a >= 12.0)')
        # LessEqual
        x = myokit.LessEqual(a, b)
        self.assertEqual(w.ex(x), '(c.a <= 12.0)')

        # Not
        cond1 = myokit.parse_expression('5 > 3')
        cond2 = myokit.parse_expression('2 < 1')
        x = myokit.Not(cond1)
        self.assertEqual(w.ex(x), '!((5.0 > 3.0))')
        # And
        x = myokit.And(cond1, cond2)
        self.assertEqual(w.ex(x), '((5.0 > 3.0) && (2.0 < 1.0))')
        # Or
        x = myokit.Or(cond1, cond2)
        self.assertEqual(w.ex(x), '((5.0 > 3.0) || (2.0 < 1.0))')

        # If (custom function)
        x = myokit.If(cond1, a, b)
        self.assertEqual(w.ex(x), 'ifthenelse((5.0 > 3.0), c.a, 12.0)')
        # Piecewise
        c = myokit.Number(1)
        x = myokit.Piecewise(cond1, a, cond2, b, c)
        self.assertEqual(
            w.ex(x),
            'ifthenelse((5.0 > 3.0), c.a, ifthenelse((2.0 < 1.0), 12.0, 1.0))')

        # Test fetching using ewriter method
        w = myokit.formats.ewriter('matlab')
        self.assertIsInstance(w, myokit.formats.matlab.MatlabExpressionWriter)

        # Test without a Myokit expression
        self.assertRaisesRegex(
            ValueError, 'Unknown expression type', w.ex, 7)


class NumPyExpressionWriterTest(unittest.TestCase):
    """ Test the NumPy ewriter class. """

    def test_all(self):
        w = myokit.formats.python.NumPyExpressionWriter()

        model = myokit.Model()
        component = model.add_component('c')
        avar = component.add_variable('a')

        # Name
        a = myokit.Name(avar)
        self.assertEqual(w.ex(a), 'c.a')
        # Number with unit
        b = myokit.Number('12', 'pF')
        self.assertEqual(w.ex(b), '12.0')

        # Prefix plus
        x = myokit.PrefixPlus(b)
        self.assertEqual(w.ex(x), '12.0')
        # Prefix minus
        x = myokit.PrefixMinus(b)
        self.assertEqual(w.ex(x), '(-12.0)')

        # Plus
        x = myokit.Plus(a, b)
        self.assertEqual(w.ex(x), 'c.a + 12.0')
        # Minus
        x = myokit.Minus(a, b)
        self.assertEqual(w.ex(x), 'c.a - 12.0')
        # Multiply
        x = myokit.Multiply(a, b)
        self.assertEqual(w.ex(x), 'c.a * 12.0')
        # Divide
        x = myokit.Divide(a, b)
        self.assertEqual(w.ex(x), 'c.a / 12.0')

        # Quotient
        x = myokit.Quotient(a, b)
        self.assertEqual(w.ex(x), 'c.a // 12.0')
        # Remainder
        x = myokit.Remainder(a, b)
        self.assertEqual(w.ex(x), 'c.a % 12.0')

        # Power
        x = myokit.Power(a, b)
        self.assertEqual(w.ex(x), 'c.a ** 12.0')
        # Sqrt
        x = myokit.Sqrt(b)
        self.assertEqual(w.ex(x), 'numpy.sqrt(12.0)')
        # Exp
        x = myokit.Exp(a)
        self.assertEqual(w.ex(x), 'numpy.exp(c.a)')
        # Log(a)
        x = myokit.Log(b)
        self.assertEqual(w.ex(x), 'numpy.log(12.0)')
        # Log(a, b)
        x = myokit.Log(a, b)
        self.assertEqual(w.ex(x), 'numpy.log(c.a, 12.0)')
        # Log10
        x = myokit.Log10(b)
        self.assertEqual(w.ex(x), 'numpy.log10(12.0)')

        # Sin
        x = myokit.Sin(b)
        self.assertEqual(w.ex(x), 'numpy.sin(12.0)')
        # Cos
        x = myokit.Cos(b)
        self.assertEqual(w.ex(x), 'numpy.cos(12.0)')
        # Tan
        x = myokit.Tan(b)
        self.assertEqual(w.ex(x), 'numpy.tan(12.0)')
        # ASin
        x = myokit.ASin(b)
        self.assertEqual(w.ex(x), 'numpy.arcsin(12.0)')
        # ACos
        x = myokit.ACos(b)
        self.assertEqual(w.ex(x), 'numpy.arccos(12.0)')
        # ATan
        x = myokit.ATan(b)
        self.assertEqual(w.ex(x), 'numpy.arctan(12.0)')

        # Floor
        x = myokit.Floor(b)
        self.assertEqual(w.ex(x), 'numpy.floor(12.0)')
        # Ceil
        x = myokit.Ceil(b)
        self.assertEqual(w.ex(x), 'numpy.ceil(12.0)')
        # Abs
        x = myokit.Abs(b)
        self.assertEqual(w.ex(x), 'abs(12.0)')

        # Equal
        x = myokit.Equal(a, b)
        self.assertEqual(w.ex(x), '(c.a == 12.0)')
        # NotEqual
        x = myokit.NotEqual(a, b)
        self.assertEqual(w.ex(x), '(c.a != 12.0)')
        # More
        x = myokit.More(a, b)
        self.assertEqual(w.ex(x), '(c.a > 12.0)')
        # Less
        x = myokit.Less(a, b)
        self.assertEqual(w.ex(x), '(c.a < 12.0)')
        # MoreEqual
        x = myokit.MoreEqual(a, b)
        self.assertEqual(w.ex(x), '(c.a >= 12.0)')
        # LessEqual
        x = myokit.LessEqual(a, b)
        self.assertEqual(w.ex(x), '(c.a <= 12.0)')

        # Not
        cond1 = myokit.parse_expression('5 > 3')
        cond2 = myokit.parse_expression('2 < 1')
        x = myokit.Not(cond1)
        self.assertEqual(w.ex(x), 'not ((5.0 > 3.0))')
        # And
        x = myokit.And(cond1, cond2)
        self.assertEqual(w.ex(x), '((5.0 > 3.0) and (2.0 < 1.0))')
        # Or
        x = myokit.Or(cond1, cond2)
        self.assertEqual(w.ex(x), '((5.0 > 3.0) or (2.0 < 1.0))')

        # If
        x = myokit.If(cond1, a, b)
        self.assertEqual(
            w.ex(x), 'numpy.select([(5.0 > 3.0)], [c.a], 12.0)')
        # Piecewise
        c = myokit.Number(1)
        x = myokit.Piecewise(cond1, a, cond2, b, c)
        self.assertEqual(
            w.ex(x),
            'numpy.select([(5.0 > 3.0), (2.0 < 1.0)], [c.a, 12.0], 1.0)')

        # Test fetching using ewriter method
        w = myokit.formats.ewriter('numpy')
        self.assertIsInstance(w, myokit.formats.python.NumPyExpressionWriter)

        # Test without a Myokit expression
        self.assertRaisesRegex(
            ValueError, 'Unknown expression type', w.ex, 7)


class PythonExpressionWriterTest(unittest.TestCase):
    """ Test the Python ewriter class. """

    def test_all(self):
        w = myokit.formats.python.PythonExpressionWriter()

        model = myokit.Model()
        component = model.add_component('c')
        avar = component.add_variable('a')

        # Name
        a = myokit.Name(avar)
        self.assertEqual(w.ex(a), 'c.a')
        # Derivative
        x = myokit.Derivative(a)
        self.assertEqual(w.ex(x), 'dot(c.a)')
        # Partial derivative
        x = myokit.PartialDerivative(a, a)
        self.assertEqual(w.ex(x), 'diff(c.a, c.a)')
        # Initial value
        x = myokit.InitialValue(a)
        self.assertEqual(w.ex(x), 'init(c.a)')

        # Number
        b = myokit.Number(3)
        self.assertEqual(w.ex(b), '3.0')
        # Number with unit
        b = myokit.Number(12, 'pF')
        self.assertEqual(w.ex(b), '12.0')

        # Prefix plus
        x = myokit.PrefixPlus(b)
        self.assertEqual(w.ex(x), '12.0')
        # Prefix minus
        x = myokit.PrefixMinus(b)
        self.assertEqual(w.ex(x), '(-12.0)')

        # Plus
        x = myokit.Plus(a, b)
        self.assertEqual(w.ex(x), 'c.a + 12.0')
        # Minus
        x = myokit.Minus(a, b)
        self.assertEqual(w.ex(x), 'c.a - 12.0')
        # Multiply
        x = myokit.Multiply(a, b)
        self.assertEqual(w.ex(x), 'c.a * 12.0')
        # Divide
        x = myokit.Divide(a, b)
        self.assertEqual(w.ex(x), 'c.a / 12.0')

        # Quotient
        x = myokit.Quotient(a, b)
        self.assertEqual(w.ex(x), 'c.a // 12.0')
        # Remainder
        x = myokit.Remainder(a, b)
        self.assertEqual(w.ex(x), 'c.a % 12.0')

        # Power
        x = myokit.Power(a, b)
        self.assertEqual(w.ex(x), 'c.a ** 12.0')
        # Sqrt
        x = myokit.Sqrt(b)
        self.assertEqual(w.ex(x), 'math.sqrt(12.0)')
        # Exp
        x = myokit.Exp(a)
        self.assertEqual(w.ex(x), 'math.exp(c.a)')
        # Log(a)
        x = myokit.Log(b)
        self.assertEqual(w.ex(x), 'math.log(12.0)')
        # Log(a, b)
        x = myokit.Log(a, b)
        self.assertEqual(w.ex(x), 'math.log(c.a, 12.0)')
        # Log10
        x = myokit.Log10(b)
        self.assertEqual(w.ex(x), 'math.log10(12.0)')

        # Sin
        x = myokit.Sin(b)
        self.assertEqual(w.ex(x), 'math.sin(12.0)')
        # Cos
        x = myokit.Cos(b)
        self.assertEqual(w.ex(x), 'math.cos(12.0)')
        # Tan
        x = myokit.Tan(b)
        self.assertEqual(w.ex(x), 'math.tan(12.0)')
        # ASin
        x = myokit.ASin(b)
        self.assertEqual(w.ex(x), 'math.asin(12.0)')
        # ACos
        x = myokit.ACos(b)
        self.assertEqual(w.ex(x), 'math.acos(12.0)')
        # ATan
        x = myokit.ATan(b)
        self.assertEqual(w.ex(x), 'math.atan(12.0)')

        # Floor
        x = myokit.Floor(b)
        self.assertEqual(w.ex(x), 'math.floor(12.0)')
        # Ceil
        x = myokit.Ceil(b)
        self.assertEqual(w.ex(x), 'math.ceil(12.0)')
        # Abs
        x = myokit.Abs(b)
        self.assertEqual(w.ex(x), 'abs(12.0)')

        # Equal
        x = myokit.Equal(a, b)
        self.assertEqual(w.ex(x), '(c.a == 12.0)')
        # NotEqual
        x = myokit.NotEqual(a, b)
        self.assertEqual(w.ex(x), '(c.a != 12.0)')
        # More
        x = myokit.More(a, b)
        self.assertEqual(w.ex(x), '(c.a > 12.0)')
        # Less
        x = myokit.Less(a, b)
        self.assertEqual(w.ex(x), '(c.a < 12.0)')
        # MoreEqual
        x = myokit.MoreEqual(a, b)
        self.assertEqual(w.ex(x), '(c.a >= 12.0)')
        # LessEqual
        x = myokit.LessEqual(a, b)
        self.assertEqual(w.ex(x), '(c.a <= 12.0)')

        # Not
        cond1 = myokit.parse_expression('5 > 3')
        cond2 = myokit.parse_expression('2 < 1')
        x = myokit.Not(cond1)
        self.assertEqual(w.ex(x), 'not ((5.0 > 3.0))')
        # And
        x = myokit.And(cond1, cond2)
        self.assertEqual(w.ex(x), '((5.0 > 3.0) and (2.0 < 1.0))')
        # Or
        x = myokit.Or(cond1, cond2)
        self.assertEqual(w.ex(x), '((5.0 > 3.0) or (2.0 < 1.0))')

        # If
        x = myokit.If(cond1, a, b)
        self.assertEqual(w.ex(x), '(c.a if (5.0 > 3.0) else 12.0)')
        # Piecewise
        c = myokit.Number(1)
        x = myokit.Piecewise(cond1, a, cond2, b, c)
        self.assertEqual(
            w.ex(x),
            '(c.a if (5.0 > 3.0) else (12.0 if (2.0 < 1.0) else 1.0))')

        # Test fetching using ewriter method
        w = myokit.formats.ewriter('python')
        self.assertIsInstance(w, myokit.formats.python.PythonExpressionWriter)

        # Test lhs method
        w.set_lhs_function(lambda x: 'sheep')
        self.assertEqual(w.ex(a), 'sheep')

        # Test without a Myokit expression
        self.assertRaisesRegex(
            ValueError, 'Unknown expression type', w.ex, 7)


class StanExpressionWriterTest(unittest.TestCase):
    """ Test the Stan ewriter class. """

    def test_all(self):
        w = myokit.formats.stan.StanExpressionWriter()

        model = myokit.Model()
        component = model.add_component('c')
        avar = component.add_variable('a')

        # Name
        a = myokit.Name(avar)
        self.assertEqual(w.ex(a), 'c.a')
        # Number with unit
        b = myokit.Number('12', 'pF')
        self.assertEqual(w.ex(b), '12.0')

        # Prefix plus
        x = myokit.PrefixPlus(b)
        self.assertEqual(w.ex(x), '12.0')
        # Prefix minus
        x = myokit.PrefixMinus(b)
        self.assertEqual(w.ex(x), '(-12.0)')

        # Plus
        x = myokit.Plus(a, b)
        self.assertEqual(w.ex(x), 'c.a + 12.0')
        # Minus
        x = myokit.Minus(a, b)
        self.assertEqual(w.ex(x), 'c.a - 12.0')
        # Multiply
        x = myokit.Multiply(a, b)
        self.assertEqual(w.ex(x), 'c.a * 12.0')
        # Divide
        x = myokit.Divide(a, b)
        self.assertEqual(w.ex(x), 'c.a / 12.0')

        # Quotient
        x = myokit.Quotient(a, b)
        self.assertEqual(w.ex(x), 'floor(c.a / 12.0)')
        # Remainder
        x = myokit.Remainder(a, b)
        self.assertEqual(w.ex(x), 'fmod(c.a, 12.0)')

        # Power
        x = myokit.Power(a, b)
        self.assertEqual(w.ex(x), 'c.a ^ 12.0')
        # Sqrt
        x = myokit.Sqrt(b)
        self.assertEqual(w.ex(x), 'sqrt(12.0)')
        # Exp
        x = myokit.Exp(a)
        self.assertEqual(w.ex(x), 'exp(c.a)')
        # Log(a)
        x = myokit.Log(b)
        self.assertEqual(w.ex(x), 'log(12.0)')
        # Log(a, b)
        x = myokit.Log(a, b)
        self.assertEqual(w.ex(x), '(log(c.a) / log(12.0))')
        # Log10
        x = myokit.Log10(b)
        self.assertEqual(w.ex(x), 'log10(12.0)')

        # Sin
        x = myokit.Sin(b)
        self.assertEqual(w.ex(x), 'sin(12.0)')
        # Cos
        x = myokit.Cos(b)
        self.assertEqual(w.ex(x), 'cos(12.0)')
        # Tan
        x = myokit.Tan(b)
        self.assertEqual(w.ex(x), 'tan(12.0)')
        # ASin
        x = myokit.ASin(b)
        self.assertEqual(w.ex(x), 'asin(12.0)')
        # ACos
        x = myokit.ACos(b)
        self.assertEqual(w.ex(x), 'acos(12.0)')
        # ATan
        x = myokit.ATan(b)
        self.assertEqual(w.ex(x), 'atan(12.0)')

        # Floor
        x = myokit.Floor(b)
        self.assertEqual(w.ex(x), 'floor(12.0)')
        # Ceil
        x = myokit.Ceil(b)
        self.assertEqual(w.ex(x), 'ceil(12.0)')
        # Abs
        x = myokit.Abs(b)
        self.assertEqual(w.ex(x), 'abs(12.0)')

        # Equal
        x = myokit.Equal(a, b)
        self.assertEqual(w.ex(x), '(c.a == 12.0)')
        # NotEqual
        x = myokit.NotEqual(a, b)
        self.assertEqual(w.ex(x), '(c.a != 12.0)')
        # More
        x = myokit.More(a, b)
        self.assertEqual(w.ex(x), '(c.a > 12.0)')
        # Less
        x = myokit.Less(a, b)
        self.assertEqual(w.ex(x), '(c.a < 12.0)')
        # MoreEqual
        x = myokit.MoreEqual(a, b)
        self.assertEqual(w.ex(x), '(c.a >= 12.0)')
        # LessEqual
        x = myokit.LessEqual(a, b)
        self.assertEqual(w.ex(x), '(c.a <= 12.0)')

        # Not
        cond1 = myokit.parse_expression('5 > 3')
        cond2 = myokit.parse_expression('2 < 1')
        x = myokit.Not(cond1)
        self.assertEqual(w.ex(x), '!((5.0 > 3.0))')
        # And
        x = myokit.And(cond1, cond2)
        self.assertEqual(w.ex(x), '((5.0 > 3.0) && (2.0 < 1.0))')
        # Or
        x = myokit.Or(cond1, cond2)
        self.assertEqual(w.ex(x), '((5.0 > 3.0) || (2.0 < 1.0))')

        # If
        x = myokit.If(cond1, a, b)
        self.assertEqual(w.ex(x), '((5.0 > 3.0) ? c.a : 12.0)')
        # Piecewise
        c = myokit.Number(1)
        x = myokit.Piecewise(cond1, a, cond2, b, c)
        self.assertEqual(
            w.ex(x),
            '((5.0 > 3.0) ? c.a : ((2.0 < 1.0) ? 12.0 : 1.0))')

        # Test fetching using ewriter method
        w = myokit.formats.ewriter('stan')
        self.assertIsInstance(w, myokit.formats.stan.StanExpressionWriter)

        # Test without a Myokit expression
        self.assertRaisesRegex(
            ValueError, 'Unknown expression type', w.ex, 7)


if __name__ == '__main__':
    unittest.main()



#
# Dependency graphing: creates graphical representations of myokit models
# Uses matplotlib.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import collections
import random

import myokit

# Don't import pyplot yet, this will crash if no window environment is loaded


def create_state_dependency_matrix(model, direct=False, knockout=None):
    """
    Creates a matrix showing state variable dependency distances.

    A distance of ``1`` from state ``x`` to state ``y`` means that ``dot(x)``
    depends on ``y``.
    A distance of ``2`` means that ``dot(x)`` does not depend on ``y``, but
    does depend on another state whose derivative depends directly on ``y``.

    State variables can be "knocked out" by adding them to the ``knockout``
    list. The rows and columns for knocked out variables will be set to zero,
    and knocked out variables won't be taken into account when calculating
    indirect dependencies (distance>1).

    Arguments:

    ``model``
        The model to create a matrix for.
    ``direct``
        True if only direct dependencies (distance=1) should be shown. In this
        case, the returned matrix will show the shape of the jacobian.
    ``knockout``
        A list of state variables or state variable names to "knock-out".

    """
    # Gather dependencies
    states = [x for x in model.states()]
    n = len(states)
    deep = model.map_deep_dependencies(collapse=True, omit_states=False)

    # Get indices to knock out, if any
    iknock = []
    if knockout is not None:
        for v in knockout:
            # Ensure all items in knockout are state variables from this model
            v = model.get(v, class_filter=myokit.Variable)
            if not v.is_state():
                raise ValueError('Knockout variables must be states.')
            iknock.append(v.index())

    # Create matrix of direct dependencies
    m = []
    for i, var in enumerate(states):
        row = [0] * n
        if i not in iknock:
            for dep in deep[var.lhs()]:
                if dep.var().is_state():
                    j = dep.var().index()
                    if j not in iknock:
                        row[j] = 1
        m.append(row)

    # Return directly
    if direct:
        return m

    # Create copy of m
    q = [list(row) for row in m]

    # The distance at step+1 will be set in q, calculated from m
    # At the end of each iteration, q is copied into m
    # At the end of this operation, m = q = the distances
    # Repeat this trick n-1 times...
    for index in range(1, n):
        for i in range(0, n):
            for j in range(0, n):
                # ... for every item
                if m[i][j] == index:
                    # ... where the value is the current index 1,2,3...
                    for k in range(0, n):
                        # ... update the entries
                        # ... unless a previous value was set
                        if q[i][k] == 0 and m[j][k] > 0:
                            q[i][k] = m[i][j] + m[j][k]
        for i in range(0, n):
            for j in range(0, n):
                m[i][j] = q[i][j]

    return m


def plot_state_dependency_matrix(
        model, direct=False, knockout=[], axes=None):
    """
    Creates a matrix showing state variable dependency distances.

    To show only direct (first-order) dependencies, set the optional argument
    ``direct`` to ``True``.

    Variables can be "knocked out" by adding them to the list in the
    ``knockout`` parameter. If x depends on y and y depends on z, knocking out
    y will prevent the method from findind x's dependency on z.

    Returns a matplotlib axes object.
    """
    import matplotlib.pyplot as plt

    # Create dependency matrix
    m = create_state_dependency_matrix(model, direct, knockout)
    n = len(m)

    # Configure axes
    a = axes if axes is not None else plt.gca()
    a.set_aspect('equal')
    a.set_xlim(0, n + 2)
    a.set_ylim(0, n)

    # Create matrix plot
    from matplotlib.patches import Rectangle
    w = 1.0 / (max(max(m)))     # color weight
    p1 = 0.1                    # Padding
    p2 = 1.0 - 2 * p1           # Square size

    def c(v):
        # Colormap
        if v == 1:
            return (0, 0, 0)
        else:
            vv = v * w
            if (vv > 1.0):
                return (0, 1.0, 0)
            return (vv, vv, 1.0)

    for y, row in enumerate(m):
        y = n - y - 1
        for x, v in enumerate(row):
            r = Rectangle((x + p1, y + p1), p2, p2, color=c(v))
            a.add_patch(r)
            r.set_clip_box(a.bbox)

        # Add colorbar
        r = Rectangle((1 + n + p1, y + p1), p2, p2, color=c(1 + y))
        a.add_patch(r)
        r.set_clip_box(a.bbox)
        a.text(
            1.5 + n,
            y + 0.5,
            str(y + 1),
            horizontalalignment='center',
            verticalalignment='center')

    # Set tick labels
    names = [i.qname() for i in model.states()]
    a.set_xticks([i + 0.5 for i in range(0, n)])
    a.set_xticklabels(names, rotation=90)
    a.xaxis.set_ticks_position('top')

    a.set_yticks([i + 0.5 for i in range(0, n)])
    rnames = list(names)
    rnames.reverse()
    a.set_yticklabels(rnames)

    # Add axes labels
    a.set_xlabel('Affecting variable')
    a.set_ylabel('Affected variable')

    # Return
    return a


class DiGraph:
    """
    A simple directed graph implementation.

    If desired, a digraph can be created from an n-by-n connectivity matrix,
    for example ``matrix=[[0, 1, 1], [0, 1, 0], [0, 0, 0]]``
    """
    def __init__(self, matrix=None):
        super().__init__()
        if isinstance(matrix, DiGraph):
            # Clone
            self.nodes = collections.OrderedDict()
            for node in matrix:
                self.add_node(node.uid)
            for node in matrix:
                for edge in node.edgo:
                    self.add_edge(node.uid, edge.uid)
        else:
            # Create new
            if matrix is None:
                matrix = []
            self.build_from_matrix(matrix)

    def __len__(self):
        return len(self.nodes)

    def __iter__(self):
        return iter(self.nodes.values())

    def __getitem__(self, key):
        return self.nodes.__getitem__(key)

    def add_edge(self, node1, node2):
        """
        Adds an edge from node1 to node2
        """
        node1 = self.uid_or_node(node1)
        node2 = self.uid_or_node(node2)
        node1.add_edge_to(node2)

    def add_node(self, node):
        """
        Adds a node. You can pass an existing ``Node`` object or an object to
        use as a new node's id.
        """
        if not isinstance(node, Node):
            node = Node(node)
        uid = node.uid
        if uid in self.nodes:
            raise ValueError('Duplicate node id: "' + str(node) + '"')
        if node.graph is not None and node.graph != self:
            raise ValueError(
                'Node already in another graph: "' + str(node) + '"')
        self.nodes[uid] = node
        node.graph = self
        return node

    def build_from_matrix(self, matrix, edges_only=False):
        """
        Replaces this graph's structure with the graph defined by the given
        n by n connectivity matrix.

        If ``edges_only`` is set to True, the matrix must have the same size
        as the current number of nodes. In this, all existing edges will be
        removed and replaced by the ones given in the connectivity matrix.
        """
        # Check matrix structure
        n = len(matrix)
        for row in matrix:
            if len(row) != n:
                raise ValueError('Matrix must be n by n')
        if edges_only:
            # Test if matrix size is same as current graph
            if n != len(self.nodes):
                raise ValueError('Matrix must have same size as graph.')
            # Clear existing edges
            for node in self.nodes.values():
                node.clear_edges()
        else:
            # Delete nodes
            self.nodes = collections.OrderedDict()
            # Create nodes
            for i in range(n):
                self.add_node(i)
        # Add edges
        nodes = list(self.nodes.values())
        for i, row in enumerate(matrix):
            for j, edge in enumerate(row):
                if edge:
                    self.add_edge(nodes[i], nodes[j])

    def cg_layers_dag(self):
        """
        Returns a layering according to the Coffman-Graham ordering scheme.

        Returns a list of lists, where each inner list represents a consecutive
        layer.

        Will raise an exception if this digraph has cycles.
        """
        # Get minimal equivalent graph
        graph = self.meg_dag()

        # Number of nodes
        n = len(graph)

        # Values of nodes
        P = collections.OrderedDict()
        for node in graph:
            P[node] = 0

        # Pick a random node with no leaving edges (successors)
        first = None
        for node in graph:
            if len(node.edgo) == 0:
                first = node
                break
        if first is None:
            raise Exception(
                'Graph is not acyclical: no node without successors found.')
        P[first] = 1

        # Assign remaining orders
        for i in range(2, n + 1):
            kmin = None
            nmin = None
            for node in graph:
                # Only test unassigned node
                if P[node] > 0:
                    continue
                # Get values of successors
                k = [0] * len(node.edgo)
                for j, s in enumerate(node.edgo):
                    p = P[s]
                    if p == 0:
                        k = None
                        break
                    k[j] = p
                # Only test nodes whose successors have all been assigned
                if k is None:
                    continue
                k.sort(reverse=True)
                if kmin is None or k < kmin:
                    kmin = k
                    nmin = node
            # Assign value to node with minimum value sequence
            P[nmin] = i

        # Assign levels
        L = []      # Layers
        D = set()   # Contains nodes already assigned
        E = set()   # Nodes on this level
        for node in sorted(P, key=P.get):
            if not node.edgo.issubset(D):
                D = D.union(E)
                L.append([self[x.uid] for x in E])
                E.clear()
            E.add(node)
        L.append([self[x.uid] for x in E])

        # Convert to lists
        layers = []
        for layer in L:
            layers.append(list(layer))
        return layers

    def meg_dag(self):
        """
        Finds a Minimal Equivalent Graph (MEG) of a Directed Acyclic Graph
        (DAG) using the algorithm by Harry Hsu [1].

        Will raise an exception if this digraph has self-referencing nodes.

        [1] An algorithm for finding a minimal equivalent graph of a digraph.
            Harry T. Hsu (1975) Journal of the Assoclatlon for Computing
            Machinery, Vol 22, No. 1, January 1975, pp 11-16
        """
        m = self.path_matrix()
        n = len(m)
        for i in range(n):
            if m[i][i]:
                raise Exception('Graph has self-referencing nodes.')
        for j in range(n):
            for i in range(n):
                if m[i][j]:
                    for k in range(n):
                        if m[j][k]:
                            m[i][k] = 0
        graph = DiGraph(self)
        graph.build_from_matrix(m, edges_only=True)
        return graph

    def matrix(self):
        """
        Returns a connectivity matrix for this graph.
        """
        nodes = list(self.nodes.values())
        n = len(nodes)
        m = [0] * n
        for i, node in enumerate(nodes):
            row = [0] * n
            for j in range(0, n):
                if node.has_edge_to(nodes[j]):
                    row[j] = 1
            m[i] = row
        return m

    def node(self, uid):
        """
        Returns the node with this id
        """
        if uid not in self.nodes:
            raise ValueError('Node not found: "' + str(uid) + '"')
        return self.nodes[uid]

    def path_matrix(self):
        """
        Returns a path matrix for this graph, showing which nodes are reachable
        from which.
        """
        p = self.matrix()
        n = len(p)
        for i in range(0, n):
            for j in range(0, n):
                if i == j:
                    continue
                if p[j][i]:
                    for k in range(0, n):
                        if p[j][k] == 0:
                            p[j][k] = p[i][k]
            return p

    def text(self, matrix=False):
        """
        Returns an ascii view of this graph
        """
        if matrix:
            m = self.matrix()
            o = []
            for row in m:
                o.append(', '.join([str(x) for x in row]))
            return '\n'.join(o)
        else:
            if self.nodes:
                out = []
                for node in self.nodes.values():
                    out.append('Node "' + str(node) + '"')
                    for edge in node.edgo:
                        out.append('  > Node "' + str(edge) + '"')
                    #out.append('')
                return '\n'.join(out)
            else:
                return 'Empty graph'

    def layout_layered(self):
        """
        Changes the x,y coordinates of this graph's node resulting in a layered
        layout.
        """
        # Get layers
        layers = self.cg_layers_dag()

        # Set initial positions
        fy = 1.0 / len(layers)
        for i, layer in enumerate(layers):
            y = fy * i
            n = len(layer)
            if n == 1:
                layer[0].x = 0.5
                layer[0].y = y
            else:
                fx = 1.0 / len(layer)
                for j, node in enumerate(layer):
                    node.y = y
                    node.x = j * fx

        # Iteratively update x coordinates using median rule
        max_runs = 100
        for n_runs in range(max_runs):
            change = False
            for i, layer in enumerate(layers):
                # Set nodes to median of neighbors
                for j, node in enumerate(layer):
                    n = len(node.edgi) + len(node.edgo)
                    if n > 0:
                        nodes = list(node.edgi)
                        nodes.extend(list(node.edgo))
                        nodes.sort(key=lambda snode: snode.x)
                        node.x = nodes[int(n / 2)].x
                # Fix possible issues with new coordinates
                org = list(layer)
                layer.sort(key=lambda snode: snode.x)
                if layer != org:
                    change = True
                # Layer-wide operations
                if i == 0:
                    # Bottom layer? Then redistribute nodes
                    fx = 1.0 / len(layer)
                    f1 = 0.8
                    f2 = 1.0 - f1
                    for k, node in enumerate(layer):
                        node.x = f1 * k * fx + f2 * node.x
                else:
                    # Resolve equal x positions
                    n = len(layer)
                    too_close = 0.1 / n
                    last = None
                    for j, node in enumerate(layer):
                        if last is None:
                            last = node
                            continue
                        if abs(node.x - last.x) < too_close:
                            eq = [last, node]
                            # Get last x
                            x1 = 0.0 if j == 1 else layer[j - 2].x
                            # Get next x
                            x2 = None
                            for k in range(j + 1, n):
                                if abs(layer[k].x - node.x) >= too_close:
                                    x2 = layer[k].x     # pragma: no cover
                                    break               # pragma: no cover
                                eq.append(layer[k])
                            if x2 is None:
                                x2 = 1.0
                            # Space nodes between x1 and x2
                            dx = (x2 - x1) / (1.0 + len(eq))
                            for k, enode in enumerate(eq):
                                enode.x = x1 + dx * (1 + k)
                        last = node
            if not change:
                break

    def remove_node(self, node):
        """
        Removes a node from this graph.
        """
        node = self.uid_or_node(node)
        node.clear_edges()
        del self.nodes[node.uid]

    def uid_or_node(self, test):
        """
        Safely turns 'test' into a node. Throws exceptions if it can't.
        """
        if not isinstance(test, Node):
            return self.node(test)
        if test.graph != self:
            raise ValueError('Node from another graph: "' + str(test) + '"')
        assert (test.uid in self.nodes and self.nodes[test.uid] == test)
        return test


class Node:
    """
    Defines a node in a graph
    """
    def __init__(self, uid):
        """
        Creates a new, graphless node with the given identifier
        """
        super().__init__()
        self.graph = None
        self.uid = uid
        self.edgo = set()
        self.edgi = set()
        # Graphical
        self.x = 0
        self.y = 0
        self.rgba = (0, 0, 1, 1)
        self.label = None

    def add_edge_to(self, node):
        """
        Ensures an edge from this node to another
        """
        node = self.graph.uid_or_node(node)
        self.edgo.add(node)
        node.edgi.add(self)

    def clear_edges(self):
        """
        Removes any edges leading from this node.
        """
        for node in self.edgo:
            node.edgi.remove(self)
        self.edgo.clear()

    def has_edge_to(self, test):
        """
        Returns true if this node points at test
        """
        test = self.graph.uid_or_node(test)
        return test in self.edgo

    def __str__(self):
        return str(self.uid)


def plot_digraph(graph, axes=None, r_node=None):
    """
    Returns a DiGraph object to a set of matplotlib axes.

    Returns a matplotlib axes object.
    """
    import matplotlib.pyplot as plt
    import matplotlib.lines as lines
    import matplotlib.patches as patches
    limits = [None, None, None, None]
    # Create axes
    ax = axes if axes is not None else plt.gca()
    ax.set_frame_on(False)
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    xmin = ymin = 1
    xmax = ymax = 0
    # Node height and width
    if r_node is None:
        w = 0.85 / len(graph)    # Just a heuristic
    else:
        w = r_node
    # Draw edges, arrows
    fArrow = 0.8
    for node in graph:
        for mode in node.edgo:
            x1, y1, x2, y2 = node.x, node.y, mode.x, mode.y
            line = lines.Line2D(
                (x1, x2), (y1, y2), lw=2, color=(0, 0, 0), alpha=.5, zorder=0)
            ax.add_line(line)
            ax.add_patch(
                patches.FancyArrowPatch(
                    (x1, y1),
                    (x1 + fArrow * (x2 - x1), y1 + fArrow * (y2 - y1)),
                    arrowstyle='->',
                    mutation_scale=60,
                    color=(0, 0, 0),
                    alpha=.5,
                    zorder=0))
        xmin = min(xmin, node.x)
        xmax = max(xmax, node.x)
        ymin = min(ymin, node.y)
        ymax = max(ymax, node.y)
    p = 0.8  # use 0.5 to show the whole node, anything higher is padding
    if limits[0] is None:
        limits[0] = xmin - p * w
    if limits[1] is None:
        limits[1] = xmax + p * w
    if limits[2] is None:
        limits[2] = ymin - p * w
    if limits[3] is None:
        limits[3] = ymax + p * w
    ax.set_xlim(limits[0], limits[1])
    ax.set_ylim(limits[2], limits[3])
    # Draw nodes
    for node in graph:
        # Draw ellipse
        e = patches.Ellipse(
            xy=(node.x, node.y),
            width=w,
            height=w,
            angle=0,
            zorder=1)
        e.set_clip_box(ax.bbox)
        e.set_alpha(node.rgba[3])
        e.set_facecolor(node.rgba[0:3])
        ax.add_artist(e)
        # Draw label
        if node.label is not None:
            ax.text(
                node.x, node.y, node.label, horizontalalignment='center',
                zorder=2)
    # Return axes
    return ax


def create_component_dependency_graph(
        model, omit_states=True, omit_constants=False):
    """
    Creates and returns a component dependency graph.
    """
    # Create graph
    g = DiGraph()
    # Add nodes
    deps = model.map_component_dependencies(
        omit_states=omit_states, omit_constants=omit_constants)
    for c in deps:
        node = g.add_node(c)
        node.x, node.y = (random.random(), random.random())
        node.rgba = (random.random(), random.random(), random.random(), 0.9)
        node.label = c.name()
    # Add edges, get sorted dict of nodes sorted by #deps
    order = {}
    for c, cdeps in deps.items():
        n = 0
        for d in cdeps:
            g.add_edge(c, d)
            n += 1
        if n not in order:
            order[n] = []
        order[n].append(c)
    # Layout
    try:
        g.layout_layered()
    except Exception:
        # Cyclical graph, create simple layout
        nOrder = len(order)
        p = 0.1
        dy = (1 - 2 * p) / (nOrder - 1) if nOrder > 1 else 0
        y = p
        for key in sorted(order.keys()):
            level = order[key]
            nLevel = len(level)
            dx = (1 - 2 * p) / (nLevel - 1) if nLevel > 1 else 0
            x = 0.5 * (1 - dx * (nLevel - 1))
            for c in level:
                node = g[c]
                node.x, node.y = (x, y)
                x += dx
            y += dy
    return g


def plot_component_dependency_graph(
        model, axes=None, omit_states=True, omit_constants=False):
    """
    Draws a graph showing the dependencies between a model's components.

    Returns a matplotlib axes object.
    """
    return plot_digraph(
        create_component_dependency_graph(
            model, omit_states=omit_states, omit_constants=omit_constants),
        axes)


def create_variable_dependency_graph(model):
    """
    Creates a dependency graph from the given model

    (Doesn't include constants)
    """
    g = DiGraph()
    # Get shallow dependencies
    deps = model.map_shallow_dependencies(
        collapse=True, omit_states=True, omit_constants=True)
    # Create nodes
    for lhs, dps in deps.items():
        # Create node
        node = g.add_node(lhs)
        # Set node label
        node.label = lhs.var().qname()
        if type(lhs) == myokit.Derivative:
            node.label = 'd(' + node.label + ')'
        # Set color
        node.rgba = (random.random(), random.random(), random.random(), 0.9)
    # Create edges
    used = set()
    for lhs, dps in deps.items():
        for dep in dps:
            g.add_edge(lhs, dep)
            used.add(lhs)
            used.add(dep)
    # Remove unused nodes
    for node in set(g.nodes).difference(used):
        g.remove_node(node)
    del used
    # Layout graphs
    g.layout_layered()
    # Return
    return g


def plot_variable_dependency_graph(model, axes=None):
    """
    Draws a graph showing the dependencies between a model's variables.

    Returns a matplotlib axes object.
    """
    return plot_digraph(
        create_variable_dependency_graph(model), axes=axes, r_node=0.03)



#
# Provides support for working with data in formats used by HEKA.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
from ._patchmaster import ( # noqa
    AmplifierMode,
    EndianAwareReader,
    Group,
    NoSupportedDAChannelError,
    PatchMasterFile,
    PulsedFile,
    Segment,
    SegmentClass,
    SegmentIncrement,
    SegmentStorage,
    Series,
    Stimulus,
    StimulusChannel,
    StimulusFile,
    Sweep,
    Trace,
    TreeNode,
)
from ._importer import PatchMasterImporter


# Importers
_importers = {
    'heka': PatchMasterImporter,
}


def importers():
    """ Returns a dict of all importers available in this module. """
    return dict(_importers)
# Exporters
# Expression writers



#!/usr/bin/env python3
#
# Tests the model building API.
#
# NOTE: THESE TESTS SHOULD SLOWLY BE MERGED INTO test_model, test_variable, etc
#
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit
import unittest

# Further model API tests are found in:
#  - test_dependency_checking.py


class ModelBuildTest(unittest.TestCase):

    def test_model_creation(self):
        # Create a model
        m = myokit.Model('LotkaVolterra')

        # Add engine component
        E = m.add_component('engine')
        self.assertEqual(len(m), 1)
        time = E.add_variable('time')
        time.set_rhs(0)
        self.assertIsNone(time.binding())
        time.set_binding('time')
        self.assertIsNotNone(time.binding())

        # Add the first test component
        X = m.add_component('X')
        self.assertNotEqual(E, X)
        self.assertEqual(X.qname(), 'X')
        self.assertEqual(X.parent(), m)
        self.assertIsInstance(X, myokit.Component)
        self.assertIn(X.qname(), m)
        self.assertEqual(len(m), 2)

        # Add variable a
        self.assertFalse(X.has_variable('a'))
        a = X.add_variable('a')
        self.assertTrue(X.has_variable('a'))
        self.assertEqual(a, a)
        self.assertIsInstance(a, myokit.Variable)
        self.assertEqual(len(X), 1)
        self.assertIn(a.name(), X)
        a.set_rhs(3)
        self.assertFalse(a.is_state())
        self.assertFalse(a.is_intermediary())
        self.assertTrue(a.is_constant())
        self.assertEqual(a.lhs(), myokit.Name(a))
        self.assertEqual(a.rhs(), myokit.Number(3))
        self.assertEqual(a.rhs().eval(), 3)
        self.assertEqual(a.code(), 'a = 3\n')
        self.assertEqual(a.eq().code(), 'X.a = 3')
        self.assertEqual(a.lhs().code(), 'X.a')
        self.assertEqual(a.rhs().code(), '3')
        self.assertEqual(
            a.eq(), myokit.Equation(myokit.Name(a), myokit.Number(3)))

        # Check lhs
        a_name1 = myokit.Name(a)
        a_name2 = myokit.Name(a)
        self.assertEqual(a_name1, a_name1)
        self.assertEqual(a_name2, a_name2)
        self.assertEqual(a_name1, a_name2)
        self.assertEqual(a_name2, a_name1)

        # Add variable b with two temporary variables
        b = X.add_variable('b')
        self.assertIsInstance(b, myokit.Variable)
        self.assertEqual(len(X), 2)
        self.assertIn(b.name(), X)
        self.assertFalse(b.has_variable('b1'))
        b1 = b.add_variable('b1')
        self.assertTrue(b.has_variable('b1'))
        self.assertEqual(len(b), 1)
        self.assertIn(b1.name(), b)
        self.assertIsInstance(b1, myokit.Variable)
        b2 = b.add_variable('b2')
        self.assertEqual(len(b), 2)
        self.assertIn(b2.name(), b)
        self.assertIsInstance(b2, myokit.Variable)
        b1.set_rhs(1)
        b2.set_rhs(
            myokit.Minus(
                myokit.Minus(myokit.Name(a), myokit.Name(b1)),
                myokit.Number(1))
        )
        b.set_rhs(myokit.Plus(myokit.Name(b1), myokit.Name(b2)))
        self.assertEqual(b.rhs().eval(), 2)
        self.assertFalse(b.is_state())
        self.assertFalse(b.is_intermediary())
        self.assertTrue(b.is_constant())
        self.assertEqual(b.lhs(), myokit.Name(b))

        # Add state variable x
        x = X.add_variable('x')
        x.set_rhs(10)
        x.promote()
        self.assertNotEqual(x, X)
        self.assertIsInstance(x, myokit.Variable)
        self.assertEqual(len(X), 3)
        self.assertIn(x.name(), X)
        self.assertTrue(x.is_state())
        self.assertFalse(x.is_intermediary())
        self.assertFalse(x.is_constant())
        self.assertEqual(x.lhs(), myokit.Derivative(myokit.Name(x)))
        self.assertEqual(x.index(), 0)

        # Test demoting, promoting
        x.demote()
        self.assertFalse(x.is_state())
        self.assertFalse(x.is_intermediary())
        self.assertTrue(x.is_constant())
        self.assertEqual(x.lhs(), myokit.Name(x))
        x.promote()
        self.assertTrue(x.is_state())
        self.assertFalse(x.is_intermediary())
        self.assertFalse(x.is_constant())
        self.assertEqual(x.lhs(), myokit.Derivative(myokit.Name(x)))
        x.demote()
        x.promote()
        x.demote()
        x.promote()
        self.assertTrue(x.is_state())
        self.assertFalse(x.is_intermediary())
        self.assertFalse(x.is_constant())
        self.assertEqual(x.lhs(), myokit.Derivative(myokit.Name(x)))

        # Set number initial value
        x.demote()
        x.promote(1)
        self.assertEqual(x.initial_value(as_float=True), 1)

        # Set constant expression initial value
        x.set_initial_value(myokit.Name(b))
        self.assertEqual(x.initial_value(as_float=True), 2)
        self.assertEqual(x.initial_value(), myokit.Name(b))

        # Non-constant expression initial value can be set, but isn't valid
        m.validate()
        b.promote()
        self.assertRaisesRegex(
            myokit.IntegrityError, 'not constant', m.validate)
        b.demote()

        # Set literal valued expression initial value
        x.demote()
        x.promote('1 + 2')
        self.assertEqual(x.initial_value(as_float=True), 3)

        # Add second component, variables
        Y = m.add_component('Y')
        self.assertNotEqual(X, Y)
        self.assertEqual(len(m), 3)
        c = Y.add_variable('c')
        c.set_rhs(myokit.Minus(myokit.Name(a), myokit.Number(1)))
        d = Y.add_variable('d')
        d.set_rhs(2)
        y = Y.add_variable('y')
        y.promote()

        # Set rhs for x and y
        x.set_rhs(myokit.Minus(
            myokit.Multiply(myokit.Name(a), myokit.Name(x)),
            myokit.Multiply(
                myokit.Multiply(myokit.Name(b), myokit.Name(x)),
                myokit.Name(y)
            )
        ))
        x.set_initial_value(10)
        self.assertEqual(x.rhs().code(), 'X.a * X.x - X.b * X.x * Y.y')
        y.set_rhs(myokit.Plus(
            myokit.Multiply(
                myokit.PrefixMinus(myokit.Name(c)), myokit.Name(y)
            ),
            myokit.Multiply(
                myokit.Multiply(myokit.Name(d), myokit.Name(x)),
                myokit.Name(y)
            )
        ))
        y.set_initial_value(5)
        self.assertEqual(y.rhs().code(), '-Y.c * Y.y + Y.d * X.x * Y.y')

        # Add ano component, variables
        Z = m.add_component('Z')
        self.assertNotEqual(X, Z)
        self.assertNotEqual(Y, Z)
        self.assertEqual(len(m), 4)
        t = Z.add_variable('total')
        self.assertEqual(t.name(), 'total')
        self.assertEqual(t.qname(), 'Z.total')
        self.assertEqual(t.qname(X), 'Z.total')
        self.assertEqual(t.qname(Z), 'total')
        t.set_rhs(myokit.Plus(myokit.Name(x), myokit.Name(y)))
        self.assertFalse(t.is_state())
        self.assertFalse(t.is_constant())
        self.assertTrue(t.is_intermediary())
        self.assertEqual(t.rhs().code(), 'X.x + Y.y')
        self.assertEqual(t.rhs().code(X), 'x + Y.y')
        self.assertEqual(t.rhs().code(Y), 'X.x + y')
        self.assertEqual(t.rhs().code(Z), 'X.x + Y.y')

        # Check state
        state = [i for i in m.states()]
        self.assertEqual(len(state), 2)
        self.assertIn(x, state)
        self.assertIn(y, state)

        # Test variable iterators
        def has(*v):
            for var in v:
                self.assertIn(var, vrs)
            self.assertEqual(len(vrs), len(v))
        vrs = [i for i in m.variables()]
        has(a, b, c, d, x, y, t, time)
        vrs = [i for i in m.variables(deep=True)]
        has(a, b, c, d, x, y, t, b1, b2, time)
        vrs = [i for i in m.variables(const=True)]
        has(a, b, c, d)
        vrs = [i for i in m.variables(const=True, deep=True)]
        has(a, b, c, d, b1, b2)
        vrs = [i for i in m.variables(const=False)]
        has(x, y, t, time)
        vrs = [i for i in m.variables(const=False, deep=True)]
        has(x, y, t, time)
        vrs = [i for i in m.variables(state=True)]
        has(x, y)
        vrs = [i for i in m.variables(state=True, deep=True)]
        has(x, y)
        vrs = [i for i in m.variables(state=False)]
        has(a, b, c, d, t, time)
        vrs = [i for i in m.variables(state=False, deep=True)]
        has(a, b, c, d, t, b1, b2, time)
        vrs = [i for i in m.variables(inter=True)]
        has(t)
        vrs = [i for i in m.variables(inter=True, deep=True)]
        has(t)
        vrs = [i for i in m.variables(inter=False)]
        has(a, b, c, d, x, y, time)
        vrs = [i for i in m.variables(inter=False, deep=True)]
        has(a, b, c, d, x, y, b1, b2, time)
        vrs = list(m.variables(const=True, state=True))
        has()
        vrs = list(m.variables(const=True, state=False))
        has(a, b, c, d)

        # Test sorted variable iteration
        names = [v.name() for v in m.variables(deep=True, sort=True)]
        self.assertEqual(names, [
            'a', 'b', 'b1', 'b2', 'x', 'c', 'd', 'y', 'total', 'time'])

        # Test equation iteration
        # Deeper testing is done when testing the ``variables`` method.
        eq = [eq for eq in X.equations(deep=False)]
        self.assertEqual(len(eq), 3)
        self.assertEqual(len(eq), X.count_equations(deep=False))
        eq = [eq for eq in X.equations(deep=True)]
        self.assertEqual(len(eq), 5)
        self.assertEqual(len(eq), X.count_equations(deep=True))
        eq = [eq for eq in Y.equations(deep=False)]
        self.assertEqual(len(eq), 3)
        self.assertEqual(len(eq), Y.count_equations(deep=False))
        eq = [eq for eq in Y.equations(deep=True)]
        self.assertEqual(len(eq), 3)
        self.assertEqual(len(eq), Y.count_equations(deep=True))
        eq = [eq for eq in Z.equations(deep=False)]
        self.assertEqual(len(eq), 1)
        self.assertEqual(len(eq), Z.count_equations(deep=False))
        eq = [eq for eq in Z.equations(deep=True)]
        self.assertEqual(len(eq), 1)
        self.assertEqual(len(eq), Z.count_equations(deep=True))
        eq = [eq for eq in E.equations(deep=False)]
        self.assertEqual(len(eq), 1)
        eq = [eq for eq in E.equations(deep=True)]
        self.assertEqual(len(eq), 1)
        eq = [eq for eq in m.equations(deep=False)]
        self.assertEqual(len(eq), 8)
        eq = [eq for eq in m.equations(deep=True)]
        self.assertEqual(len(eq), 10)

        # Test dependency mapping
        def has(var, *dps):
            lst = vrs[m.get(var).lhs() if isinstance(var, str) else var]
            self.assertEqual(len(lst), len(dps))
            for d in dps:
                d = m.get(d).lhs() if isinstance(d, str) else d
                self.assertIn(d, lst)

        vrs = m.map_shallow_dependencies(omit_states=False)
        self.assertEqual(len(vrs), 12)
        has('X.a')
        has('X.b', 'X.b.b1', 'X.b.b2')
        has('X.b.b1')
        has('X.b.b2', 'X.a', 'X.b.b1')
        has('X.x', 'X.a', 'X.b', myokit.Name(x), myokit.Name(y))
        has(myokit.Name(x))
        has('Y.c', 'X.a')
        has('Y.d')
        has('Y.y', 'Y.c', 'Y.d', myokit.Name(x), myokit.Name(y))
        has(myokit.Name(y))
        has('Z.total', myokit.Name(x), myokit.Name(y))
        vrs = m.map_shallow_dependencies()
        self.assertEqual(len(vrs), 10)
        has('X.a')
        has('X.b', 'X.b.b1', 'X.b.b2')
        has('X.b.b1')
        has('X.b.b2', 'X.a', 'X.b.b1')
        has('X.x', 'X.a', 'X.b')
        has('Y.c', 'X.a')
        has('Y.d')
        has('Y.y', 'Y.c', 'Y.d')
        has('Z.total')
        vrs = m.map_shallow_dependencies(collapse=True)
        self.assertEqual(len(vrs), 8)
        has('X.a')
        has('X.b', 'X.a')
        has('X.x', 'X.a', 'X.b')
        has('Y.c', 'X.a')
        has('Y.d')
        has('Y.y', 'Y.c', 'Y.d')
        has('Z.total')

        # Validate
        m.validate()

        # Get solvable order
        order = m.solvable_order()
        self.assertEqual(len(order), 5)
        self.assertIn('*remaining*', order)
        self.assertIn('X', order)
        self.assertIn('Y', order)
        self.assertIn('Z', order)

        # Check that X comes before Y
        pos = dict([(name, k) for k, name in enumerate(order)])
        self.assertLess(pos['X'], pos['Y'])
        self.assertEqual(pos['*remaining*'], 4)

        # Check component equation lists
        eqs = order['*remaining*']
        self.assertEqual(len(eqs), 0)
        eqs = order['Z']
        self.assertEqual(len(eqs), 1)
        self.assertEqual(eqs[0].code(), 'Z.total = X.x + Y.y')
        eqs = order['Y']
        self.assertEqual(len(eqs), 3)
        self.assertEqual(
            eqs[2].code(), 'dot(Y.y) = -Y.c * Y.y + Y.d * X.x * Y.y')
        eqs = order['X']
        self.assertEqual(len(eqs), 5)
        self.assertEqual(eqs[0].code(), 'X.a = 3')
        self.assertEqual(eqs[1].code(), 'b1 = 1')
        self.assertEqual(eqs[2].code(), 'b2 = X.a - b1 - 1')
        self.assertEqual(eqs[3].code(), 'X.b = b1 + b2')

        # Test model export and cloning
        code1 = m.code()
        code2 = m.clone().code()
        self.assertEqual(code1, code2)

    def test_scope(self):
        # Test if illegal references are detected.

        m = myokit.Model('Scope')
        c = m.add_component('c')
        p = c.add_variable('p')
        q = p.add_variable('q')
        p.set_rhs('10 + q')
        q.set_rhs('18 / 2.4')
        d = m.add_component('d')
        r = d.add_variable('r')
        r.set_rhs(myokit.Name(q))
        self.assertRaises(myokit.IllegalReferenceError, r.validate)

    def test_invalid_names(self):
        # Test if invalid or duplicate names are detected.

        m = myokit.Model('Duplicates')
        c0 = m.add_component('c0')
        t = c0.add_variable('time')
        t.set_rhs(myokit.Number(0))
        t.set_binding('time')

        # Test add component
        # Duplicates
        self.assertRaises(myokit.DuplicateName, m.add_component, 'c0')
        # Badly formed names
        self.assertRaises(myokit.InvalidNameError, m.add_component, '0')
        self.assertRaises(myokit.InvalidNameError, m.add_component, '_0')
        self.assertRaises(myokit.InvalidNameError, m.add_component, '123abvc')
        self.assertRaises(myokit.InvalidNameError, m.add_component, 'ab cd')
        self.assertRaises(myokit.InvalidNameError, m.add_component, 'ab.cd')
        self.assertRaises(myokit.InvalidNameError, m.add_component, 'ab!cd')
        self.assertRaises(myokit.InvalidNameError, m.add_component, '5*x')
        # Keywords
        self.assertRaises(myokit.InvalidNameError, m.add_component, 'and')
        self.assertRaises(myokit.InvalidNameError, m.add_component, 'bind')
        # Test adding variable to component
        c1 = m.add_component('c1')
        # Duplicate
        # Badly formed names
        self.assertRaises(myokit.InvalidNameError, c0.add_variable, '0')
        self.assertRaises(myokit.InvalidNameError, c0.add_variable, '_aap')
        self.assertRaises(myokit.InvalidNameError, c0.add_variable, '123abvc')
        self.assertRaises(myokit.InvalidNameError, c0.add_variable, 'ab cd')
        self.assertRaises(myokit.InvalidNameError, c0.add_variable, 'ab.cd')
        self.assertRaises(myokit.InvalidNameError, c0.add_variable, 'ab!cd')
        self.assertRaises(myokit.InvalidNameError, c0.add_variable, '5*x')
        # Keywords
        self.assertRaises(myokit.InvalidNameError, c0.add_variable, 'or')
        self.assertRaises(myokit.InvalidNameError, c0.add_variable, 'label')

        # Test adding variable to variable
        v1 = c0.add_variable('c0')
        # Duplicate
        v2 = c1.add_variable('c0')
        self.assertRaises(myokit.DuplicateName, c1.add_variable, 'c0')
        self.assertRaises(myokit.DuplicateName, v1.add_variable, 'c0')
        self.assertRaises(myokit.DuplicateName, v2.add_variable, 'c0')
        # Badly formed names
        self.assertRaises(myokit.InvalidNameError, v1.add_variable, '0')
        self.assertRaises(myokit.InvalidNameError, v1.add_variable, '_aap')
        self.assertRaises(myokit.InvalidNameError, v1.add_variable, '123abvc')
        self.assertRaises(myokit.InvalidNameError, v1.add_variable, 'ab cd')
        self.assertRaises(myokit.InvalidNameError, v1.add_variable, 'ab.cd')
        self.assertRaises(myokit.InvalidNameError, v1.add_variable, 'ab!cd')
        self.assertRaises(myokit.InvalidNameError, v1.add_variable, '5*x')
        # Keywords
        self.assertRaises(myokit.InvalidNameError, v1.add_variable, 'not')
        self.assertRaises(myokit.InvalidNameError, v1.add_variable, 'in')


if __name__ == '__main__':
    unittest.main()
